# cython: language_level=3
from __future__ import division
import numpy as np
import timeit
import itertools
import time
import pickle
import copy
from sys import stdout
from libc.math cimport log,exp,sqrt 
#from scipy.interpolate import interp2d,RectBivariateSpline
from scipy.special import logsumexp

class single_detector(object):
 
    def __init__(self, tr, obs, prog = False, lookup_table=None, make_vitmap = True):
        '''
        initialising viterbi class
        '''
        self.prog = prog
        
        # make sure input data of right type
        tr = np.array(tr).astype('double')
        obs = np.array(obs).astype('double')

        # make sure input data of right shape
        if len(np.shape(tr)) != 1:
            raise Exception("transition matrix wrong shape, should be Kx1 array")
        if len(np.shape(obs)) != 2:
            raise Exception("observation wrong shape, should be NxM array")
        
        self.run(tr, obs, lookup_table)
        if make_vitmap:
            self.get_vitmap()

        
    #------------------------------------------------------------------------------------------------
    # Get Viterbi map
    #------------------------------------------------------------------------------------------------

    def get_vitmap(self,log=False):
        """
        normalise the viterbi output such that the sum of eah column is 1
        """
        path_m = []
        for i in self.V:
            sump = logsumexp(i)
            val = np.exp(i-sump)
            path_m.append(val)
        path_m = np.array(path_m)
        if log:
            path_m = np.log(path_m)

        self.vitmap = path_m

    #------------------------
    # Read lookup table
    #--------------------------
    
    def read_lookup(self,lookup_table):
        """
        Read the lookup table for chosen statistic
        """
        if isinstance(lookup_table, str):
            with open(lookup_table,'rb') as f1:
                likelihood, params = pickle.load(f1)
            ranges = np.linspace(params["power_ranges"][0],params["power_ranges"][1],int(params["power_ranges"][2]))
            fact = abs(1./(ranges[0] - ranges[1]))
        else:
            # if it is a line aware stat object then creat ranges from the object
            # get the log odds ratio as the statitsic to use 
            likelihood = np.log(lookup_table.signoiseline)
            ranges = lookup_table.powers
            fact = abs(1./(ranges[0] - ranges[1]))
        
        return likelihood, ranges, fact
    
    
    #-------------------------------------------------------------------------------------------------
    # Basic Viterbi Algorithm
    #-------------------------------------------------------------------------------------------------

    
    def run(self,double[:] tr, double[:, :] obs, lookup_table = None):
        '''
        Run the viterbi algorithm for given single set of data and 3x1 transition matrix.
        This returns the track through the data which gives the largest sum of power.
        Args
        ------
        tr: array
            transition matrix of sixe Kx1
        obs: array
            observation data of size NxM
        Returns
        -------
        vit_track: array
            index of the viterbi path found in data
        max_end_prob: float
            maximum probability at end of path
        V: array
           viterbi matrix
        prev: array
            previous positions of paths for each bin
        '''
        
        # find shape of the observation and create empty array for citerbi matrix and previous track positions
        shape = np.shape(obs)
        cdef double[:, :] V = np.zeros(shape)
        cdef int[:, :] prev = np.zeros(shape,dtype=np.int32)
        
        # defining variables
        cdef int t,i,j         # indexes
        cdef double pbar = 0   # progress bar
        cdef double pt
        
        # finding length and width of observation
        cdef int length = shape[0]#len(obs)
        cdef int width = shape[1]#len(obs[0])
        
        # find size of the transition matrix
        cdef int tr_length = len(tr)
        # find half width of transntion, i.e. number of bins up and down it can move
        cdef int tr_width = int((tr_length-1)/2 )

        # vectors to store lookup tables
        cdef double[:] logarr
        cdef double fact
        cdef int[:,:] obs_ind
        if lookup_table is not None:
            # get log statistic, the ranges of lookuptable and spacing factor
            logarr, ranges, fact = self.read_lookup(lookup_table)
            # shift observation to index for array
            obs_ind_arr = (obs - ranges[0])*fact
            obs_ind_arr[obs_ind_arr >= len(ranges)] = len(ranges)-1
            obs_ind_arr[obs_ind_arr < 0] = 0
            # save array to vector
            obs_ind = np.array(obs_ind_arr).astype(np.int32)
            del obs_ind_arr
        else:
            # if using the sum of power, set element of index to 2d array
            obs_ind = np.arange(np.prod(shape)).reshape(shape).astype(np.int32)
            # flatten input spectrogram for access in 1d
            logarr = np.ravel(obs)
        
        # run for first time index, i.e. fill with observation
        for i in range(width):
            V[0][i] = logarr[obs_ind[0][i]]
            prev[0][i] = 1
        
        # run iterative part of algorithm
        for t in range(1,length):
            pt = t/(length)*100
            for i in range(width):
                temp = -1e6
                for j in range(tr_length):
                    if i+j-tr_width>=0 and i+j-tr_width<=width-1:
                        o = obs_ind[t][i]
                        value = tr[j] + logarr[o] + V[t-1][i+j-tr_width]

                        if value>temp:
                            temp = value
                            V[t][i] = temp
                            prev[t][i] = i+j-tr_width
                        elif value == temp and j == int(tr_length/2.):
                            temp = value
                            V[t][i] = temp
                            prev[t][i] = i+j-tr_width
                            

            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./len(obs)
    
        cdef double max_end_prob = max(V[length-1][i] for i in range(width))
        cdef int previous
        cdef int[:] vit_track = np.zeros(length,dtype=np.int32)
        
                        
        for i in range(width):
            if V[length-1][i] == max_end_prob:
                vit_track[length-1] = i # appends maximum path value from final step
                previous = prev[length-1][i]
                break

        for t in range(len(V)-2,-1,-1):
            vit_track[t] = previous # insert previous step
            previous = prev[t][previous]

        self.vit_track = np.array(vit_track)
        self.max_end_prob = max_end_prob
        self.V = np.array(V)
        self.prev = np.array(prev)
        

class two_detector(object):
 
    def __init__(self, tr, obs1, obs2, lookup_table_2det = None, lookup_table_1det = None, prog = False,fractions=None, make_vitmap=True):
        '''
        viterbi algorithm for two detectors with an optional line aware statistic
        uses S = \frac{p(ss \mid x)}{p(nn \mid x)+p(ns \mid x)+p(sn \mid x)}
        [currently configured for averaging over 48 SFT need to change to correctly use any]
        Args
        ---------
        tr1: array
            transition matrix
        obs11: array
            observation from detector 1 (float64 array)
        obs21: array
            observation from detector 2 (float64 array)
        lookup_table_1det: string (optional)
            filepath of lookup table for the one detector case or LineAwareStatistic claess
        lookup_table_2det: string ( optional)
            filepath of lookup table for the two detector caseor LineAwareStatistic claess 
        prog: bool (optional)
            show progress bar if True 
        fractions: array (optional)
            array of the ratio of the noise floor and amount of data between detectors
        returns
        -----------
        vit_track1: array
            path in detector 1
        vit_track2: array
            path in detector2
        vit_track_ref: array
            path in reference detector
        V: array
            viterbi values in reference detector
        prev: array
            previous positions before jump
        max_prob: double
            maximum end path probability

        '''

        if len(np.shape(obs1)) < 2 or len(np.shape(obs1)) < 2:
            raise Exception("Please input 2D array for observation data")
        if np.shape(obs1) != np.shape(obs2):
            raise Exception("Please make sure teh two input data have the same dimensions")

        tr = np.array(tr).astype('double')
        obs1 = np.array(obs1).astype('double')
        obs2 = np.array(obs2).astype('double')
        self.prog = prog

        if fractions is not None and lookup_table_2det is not None:
            self.run_lookup_amp(tr, obs1, obs2,lookup_table=lookup_table_2det,fractions=fractions)
        else:
            self.run(tr, obs1, obs2,lookup_table_1det=lookup_table_1det, lookup_table_2det=lookup_table_2det)
            
        self.get_track()
        if make_vitmap:
            self.get_vitmap()

    #------------------------------------------------------------------------------------------------
    # Get Viterbi map
    #------------------------------------------------------------------------------------------------

    def get_vitmap(self,log=False):
        """
        normalise the viterbi output such that the sum of eah column is 1
        """
        path_m = []
        for i in self.V:
            sump = logsumexp(i)
            val = np.exp(i-sump)
            path_m.append(val)
        path_m = np.array(path_m)
        if log:
            path_m = np.log(path_m)

        self.vitmap = path_m


    #--------------------------------------------------------
    # Read lookup tables
    #--------------------------------------------------------

    def read_lookup(self,lookup_table):
        """
        Read a lookup table from a file of a LineAwareStatistic object
        args 
        ------
        lookup_table: string or LineAwareStatistic
            filename to saved lookupt table or lineaware statistic object
        returns
        ---------
        likelihood: array
            returns odds ratio from file to use as statistic
        ranges: array
            ranges of the x,y axis of the odds ratio
        fact: float
            spacing of the ranges
        """

        if isinstance(lookup_table, str):
            # if it is a filename the load the file and create ranges from that
            with open(lookup_table,'rb') as f1:
                likelihood, params = pickle.load(f1)

            ranges = np.linspace(params["power_ranges"][0],params["power_ranges"][1],int(params["power_ranges"][2]))
            fact = abs(1./(ranges[0] - ranges[1]))
        else:
            # if it is a line aware stat object then creat ranges from the object
            # get the log odds ratio as the statitsic to use 
            likelihood = np.log(lookup_table.signoiseline)
            ranges = lookup_table.powers
            fact = abs(1./(ranges[0] - ranges[1]))
        
        return likelihood, ranges, fact

    def read_lookup_amp(self,lookup_table):
        """
        Read a lookup table from a file of a LineAwareStatisticAmplitude object
        args 
        ------
        lookup_table: string or LineAwareStatisticAmplitude
            filename to saved lookupt table or lineaware statistic object
        returns
        ---------
        likelihood: array
            returns odds ratio from file to use as statistic
        pow_ranges: array
            ranges of the x,y axis of the odds ratio (spectrogram power)
        frac_ranges: array
            ranges of the x,y axis of the odds ratio ()
        pow_fact: float
            spacing of the ranges of power
        frac_fact: float
            spacing of the ranges of data fraction
        """
        if isinstance(lookup_table, str):  
            with open(lookup_table,'rb') as f1:
                likelihood, params = pickle.load(f1)

            pow_ranges = np.linspace(params["power_ranges"][0],params["power_ranges"][1],int(params["power_ranges"][2]))
            frac_ranges = np.linspace(params["fraction_ranges"][0],params["fraction_ranges"][1],int(params["fraction_ranges"][2]))
            pow_fact = abs(1./(pow_ranges[0] - pow_ranges[1]))
            frac_fact = abs(1./(frac_ranges[0] - frac_ranges[1]))
        else:
            # if it is a line aware stat object then creat ranges from the object
            # get the log odds ratio as the statitsic to use 
            likelihood = np.log(lookup_table.signoiseline)
            pow_ranges = lookup_table.powers
            frac_ranges = lookup_table.fractions
            pow_fact = abs(1./(pow_ranges[0] - pow_ranges[1]))
            frac_fact = abs(1./(frac_ranges[0] - frac_ranges[1]))
        
        return likelihood.T, pow_ranges, frac_ranges, pow_fact, frac_fact

    
    #---------------------------------------------------------
    # Two detector viterbi
    #--------------------------------------------------------

    def run(self,transition, obs1, obs2, lookup_table_1det = None,lookup_table_2det = None):
        """
        viterbi algorithm for two detectors with an lookup table for the line veto statistic

        Where there is only one detectors data use the single detector search with the same statistic
        Args
        ---------
        tr1: array
            transition matrix
        obs11: array
            observation from detector 1
        obs21: array
            observation from detector 2 
        lookup_table: float
            links to a file which contains a 2d lookup table
        returns
        -----------
        V: array
            viterbi values in reference detector
        prev: array
            previous positions before jump
        """
        
        cdef double[:, :, :] tr = transition
        #cdef double[:, :] obs1 = obs11
        #cdef double[:, :] obs2 = obs21
        
        shape = np.shape(obs1)
        
        cdef int length = min(len(obs1),len(obs2))
        cdef int width = len(obs1[0])
        
        cdef double[:, :] val = np.ones(shape)*-1e6
        cdef int[:, :] prev = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det1 = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det2 = np.zeros(shape,dtype = np.int32)
        
        cdef int tr_len = len(tr)
        cdef int sep_len = len(tr[0])
        cdef int tr_dist = int(len(tr)/2. - 0.5)
        cdef int sep_dist = int(len(tr)/2. - 0.5)
        
        cdef int i
        cdef int j
        cdef int k
        cdef int m
        cdef int t
        cdef int x
        cdef int o1
        cdef int o2

        
        cdef long max_pos
        cdef long[:] indicies
        cdef double max_val
        cdef double value
        cdef double temp
        cdef int edge_range = 0

        # initialise and load lookuptable files for both line aware statistics
        cdef double[:,:] logarr_2d
        cdef double[:] logarr_1d
        cdef double fact_1d
        cdef double fact_2d
        cdef int[:,:] obs1_ind_1d
        cdef int[:,:] obs1_ind_2d

        obs1data = np.array([len(np.unique(obs1seg)) for obs1seg in obs1]) != 1
        obs2data = np.array([len(np.unique(obs2seg)) for obs2seg in obs2]) != 1
        
        if lookup_table_1det is not None:
            logarr_1d, ranges_1d, fact_1d = self.read_lookup(lookup_table_1det)
            
            obs1_ind_arr_1d = (obs1 - ranges_1d[0])*fact_1d
            obs1_ind_arr_1d[obs1_ind_arr_1d >= len(ranges_1d)] = len(ranges_1d)-1
            obs1_ind_arr_1d[obs1_ind_arr_1d < 0] = 0
            obs1_ind_1d = np.array(obs1_ind_arr_1d).astype(np.int32)
            del obs1_ind_arr_1d

            obs2_ind_arr_1d = (obs2-ranges_1d[0])*fact_1d
            obs2_ind_arr_1d[obs2_ind_arr_1d >= len(ranges_1d)] = len(ranges_1d)-1
            obs2_ind_arr_1d[obs2_ind_arr_1d < 0] = 0
            obs2_ind_1d = np.array(obs2_ind_arr_1d).astype(np.int32)
            del obs2_ind_arr_1d

        if lookup_table_2det is not None:
            logarr, ranges_2d, fact_2d = self.read_lookup(lookup_table_2det)
            diag = copy.copy(np.diag(logarr))
            logarr_2d = logarr

            # normalise the input powers to the ranges of lookup table so they are now indices
            obs1_ind_arr_2d = (obs1 - ranges_2d[0])*fact_2d
            # set powers outside range to end index
            obs1_ind_arr_2d[obs1_ind_arr_2d >= len(ranges_2d)] = len(ranges_2d)-1
            obs1_ind_arr_2d[obs1_ind_arr_2d < 0] = 0
            obs1_ind_2d = np.array(obs1_ind_arr_2d).astype(np.int32)

            # same for second detector
            obs2_ind_arr_2d = (obs2-ranges_2d[0])*fact_2d
            obs2_ind_arr_2d[obs2_ind_arr_2d >= len(ranges_2d)] = len(ranges_2d)-1
            obs2_ind_arr_2d[obs2_ind_arr_2d < 0] = 0
            obs2_ind_2d = np.array(obs2_ind_arr_2d).astype(np.int32)

            if lookup_table_1det is None:
                obs1_ind_1d = np.array(obs1_ind_arr_2d).astype(np.int32)
                obs2_ind_1d = np.array(obs2_ind_arr_2d).astype(np.int32)
                logarr_1d = diag
            
            del diag    
            del obs2_ind_arr_2d
            del obs1_ind_arr_2d

        sumv = False
	# if no lookup tables provided then use the sum of the powers
        if lookup_table_1det is None and lookup_table_2det is None:
            # if using the sum of power, set element of index to 2d array
            obs1_ind_1d = np.arange(np.prod(shape)).reshape(shape).astype(np.int32)
            obs1_ind_2d = np.arange(np.prod(shape)).reshape(shape).astype(np.int32)
            obs2_ind_1d = np.arange(np.prod(shape)).reshape(shape).astype(np.int32) + np.prod(shape)
            obs2_ind_2d = np.arange(np.prod(shape)).reshape(shape).astype(np.int32) + np.prod(shape)
            # flatten input spectrogram for access in 1d
            logarr_1d = np.ravel([obs1,obs2])
            sumv = True
            #logarr_2d = np.ravel([obs1,obs2])

        # initial viterbi loop for first time segment
        for i in range(width):
            temp = -1e6
            if obs1data[0] and obs2data[0]:
                # if data in both detectors
                for k in range(sep_len): 
                    if i+k-sep_dist>=0+edge_range and i+k-sep_dist<width-edge_range:
                        for m in range(sep_len):
                            if i+m-sep_dist>=0+edge_range and i+m-sep_dist<width-edge_range:
                                if sumv:
                                    value = tr[1][k][m] + obs1[0][i+k-sep_dist] + obs2[0][i+k-sep_dist]
                                else:
                                    o1 = obs1_ind_2d[0][i+k-sep_dist]
                                    o2 = obs2_ind_2d[0][i+m-sep_dist]
                                    value = tr[1][k][m] + logarr_2d[o1,o2]
                                if value > temp:
                                    temp = value
                                    val[0][i] = temp
                                    prev[0][i] =  i
                                    det1[0][i] =  i+k-1 
                                    det2[0][i] =  i+m-1
                                else:
                                    continue

            elif obs1data[0] and not obs2data[0]:
                o1 = obs1_ind_1d[0][i]
                k,m = 1,1
                value = tr[1][k][m] + logarr_1d[o1]
            elif obs2data[0] and not obs1data[0]:
                o2 = obs2_ind_1d[0][i]
                k,m = 1,1
                value = tr[1][k][m] + logarr_1d[o2]
            elif not obs2data[0] and not obs1data[0]:
                o1 = obs1_ind_2d[0][i]
                k,m = 1,1
                value = tr[1][k][m] + logarr_1d[o1]

                
            if value > temp:
                temp = value
                val[0][i] = temp
                prev[0][i] =  i
                det1[0][i] =  i+k-1 
                det2[0][i] =  i+m-1
            else:
                continue

        # i is current frequency position
        # j is transition in reference detector
        # k is separation of first detector
        # m is separation of second detector

        # run main loop
        pbar = 0 # progressbar
        for t in range(1,length):
            pt = t/(float(length))*100. # progress bar status
            for i in range(width):
                temp = -np.inf
                for j in range(tr_len):
                    if i+j-tr_dist>=0 and i+j-tr_dist<width:
                        if obs1data[t] and obs2data[t]:
                            for k in range(sep_len):
                                if i+k-sep_dist>=0+edge_range and i+k-sep_dist<width-edge_range:
                                    for m in range(sep_len):
                                        if i+m-sep_dist>=0+edge_range and i+m-sep_dist<width-edge_range:
                                            if sumv:
                                                value = tr[j][k][m] + obs1[t][i+k-sep_dist] + obs2[t][i+k-sep_dist] + val[t-1][i+j-tr_dist]
                                            else:
                                                o1 = obs1_ind_2d[t][i+k-sep_dist]
                                                o2 = obs2_ind_2d[t][i+m-sep_dist]
                                                value = tr[j][k][m] + logarr_2d[o1,o2] + val[t-1][i+j-tr_dist]

                                            if value > temp:
                                                temp = value
                                                val[t][i] = temp
                                                prev[t][i] =  i+j-tr_dist
                                                det1[t][i] =  i+k-sep_dist
                                                det2[t][i] =  i+m-sep_dist
                                            else:
                                                continue
                        else:
                            if obs1data[t] and not obs2data[t]:
                                o1 = obs1_ind_1d[t][i]
                                k,m = 1,1
                                value = tr[j][k][m] + logarr_1d[o1] + val[t-1][i+j-tr_dist]
                            elif obs2data[t] and not obs1data[t]:
                                o2 = obs2_ind_1d[t][i]
                                k,m = 1,1
                                value = tr[j][k][m] + logarr_1d[o2] + val[t-1][i+j-tr_dist]
                            elif not obs2data[t] and not obs1data[t]:
                                o1 = obs1_ind_2d[t][i]
                                k,m = 1,1
                                value = tr[j][k][m] + logarr_1d[o1] + val[t-1][i+j-tr_dist]

                            
                            if value > temp:
                                temp = value
                                val[t][i] = temp
                                prev[t][i] =  i+j-tr_dist
                                det1[t][i] =  i+k-sep_dist
                                det2[t][i] =  i+m-sep_dist
                            else:
                                continue

            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./length

        self.V = np.array(val)
        self.prev = np.array(prev)
        self.det1 = np.array(det1)
        self.det2 = np.array(det2)
        self.width = width
        self.length = length

    def run_lookup_amp(self,tr1, obs11, obs21,lookup_table = None, fractions = None):
        """
        viterbi algorithm for two detectors with an lookup table for the line veto statistic
        Args
        ---------
        tr1: array
            transition matrix
        obs11: array
            observation from detector 1
        obs21: array
            observation from detector 2 
        lookup_table: float
            links to a file which contains a 2d lookup table or LineAwareStatistic class
        returns
        -----------
        V: array
            viterbi values in reference detector
        prev: array
            previous positions before jump
        """
        cdef double[:, :, :] tr = tr1
        cdef double[:, :] obs1 = obs11
        cdef double[:, :] obs2 = obs21
        
        shape = np.shape(obs1)
        
        cdef int length = min(len(obs1),len(obs2))
        cdef int width = len(obs1[0])
        
        cdef double[:, :] val = np.ones(shape)*-1e6
        cdef int[:, :] prev = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det1 = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det2 = np.zeros(shape,dtype = np.int32)
        
        cdef int tr_len = len(tr)
        cdef int sep_len = len(tr[0])
        cdef int tr_dist = int(len(tr)/2. - 0.5)
        cdef int sep_dist = int(len(tr)/2. - 0.5)
        
        cdef int i
        cdef int j
        cdef int k
        cdef int m
        cdef int t
        cdef int x

        cdef int o1
        cdef int o2
        
        cdef long max_pos
        cdef long[:] indicies
        cdef double max_val
        cdef double value
        cdef double temp
        cdef int edge_range = 0


        cdef double fact
        cdef double fact_frac

        likelihood, pow_ranges, frac_ranges, pow_fact, frac_fact = self.read_lookup_amp(lookup_table)

        # define the likelihood as two array one which is the transpose of the other, such that the fractions of time one detector can always be compared with a value less than 1
        shp = np.shape(likelihood)
        logarr1 = np.zeros((2,shp[0],shp[1],shp[2]))
        logarr1[0] = likelihood
        logarr1[1] = ([np.array(lkl).T for lkl in likelihood])
        cdef double[:,:,:,:] logarr = logarr1

        # if the fractions are above a value of 1, inverse the lookup table so that only values between 0 and 1 have to be generated.
        if len(np.shape(fractions)) == 1:
            fractions = np.array([np.ones(width)*i for i in fractions])
            
        det_order1 = np.ones(np.shape(fractions)).astype(np.int32)
        det_order1[fractions > 1] = int(1)
        det_order1[fractions <= 1] = int(0)
        cdef int[:,:] det_order = det_order1
        del det_order1
        
        fractions[fractions > 1] = 1./fractions[fractions > 1]
        
        # make sure the fractions are the same dimensions as the input data and convert into an index
        frac_ind1 = np.array((fractions-frac_ranges[0])*frac_fact).astype(np.int32)
        frac_ind1[frac_ind1 < 0] = 0
        frac_ind1[frac_ind1 >= len(frac_ranges)] = -1
        cdef int[:,:] frac_ind = frac_ind1
        del frac_ind1
        
        obs1_ind_arr = (obs11 - pow_ranges[0])*pow_fact
        obs1_ind_arr[obs1_ind_arr >= len(pow_ranges)] = -1
        obs1_ind_arr[obs1_ind_arr < 0] = 0
        cdef int[:,:] obs1_ind = np.array(obs1_ind_arr).astype(np.int32)
        del obs1_ind_arr
        
        obs2_ind_arr = (obs21-pow_ranges[0])*pow_fact
        obs2_ind_arr[obs2_ind_arr >= len(pow_ranges)] = -1
        obs2_ind_arr[obs2_ind_arr < 0] = 0
        cdef int[:,:] obs2_ind = np.array(obs2_ind_arr).astype(np.int32)
        del obs2_ind_arr
        
        for i in range(width):
            temp = -1e6
            for k in range(sep_len): 
                if i+k-sep_dist>=0+edge_range and i+k-sep_dist<width-edge_range:
                    for m in range(sep_len):
                        if i+m-sep_dist>=0+edge_range and i+m-sep_dist<width-edge_range:
                            o1 = int(obs1_ind[0][i+k-sep_dist])
                            o2 = int(obs2_ind[0][i+m-sep_dist])
                            value = tr[1][k][m] + logarr[det_order[0,i],frac_ind[0,i],o1,o2]
                            if value > temp:
                                temp = value
                                val[0][i] = temp
                                prev[0][i] =  i
                                det1[0][i] =  i+k-1 
                                det2[0][i] =  i+m-1
                            else:
                                continue

        # i is current frequency position
        # j is transition in reference detector
        # k is separation of first detector
        # m is separation of second detector


        
        pbar = 0
        for t in range(1,length):
            pt = t/(float(length))*100.
            for i in range(width):
                temp = -np.inf
                for j in range(tr_len):
                    if i+j-tr_dist>=0 and i+j-tr_dist<width:
                       for k in range(sep_len):
                            if i+k-sep_dist>=0+edge_range and i+k-sep_dist<width-edge_range:
                                for m in range(sep_len):
                                    if i+m-sep_dist>=0+edge_range and i+m-sep_dist<width-edge_range:
                                        o1 = int(obs1_ind[t][i+k-sep_dist])
                                        o2 = int(obs2_ind[t][i+m-sep_dist])
                                        value = tr[j][k][m] + val[t-1][i+j-tr_dist] + logarr[det_order[t,i],frac_ind[t,i],o1,o2]

                                        if value > temp:
                                            temp = value
                                            val[t][i] = temp
                                            prev[t][i] =  i+j-tr_dist
                                            det1[t][i] =  i+k-sep_dist
                                            det2[t][i] =  i+m-sep_dist
                                        else:
                                            continue
            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./length

        self.V = np.array(val)
        self.prev = np.array(prev)
        self.det1 = np.array(det1)
        self.det2 = np.array(det2)
        self.width = width
        self.length = length

    def get_track(self):
        """
        Takes in the viterbi data from a run and identifies the track thorugh whcih gives the maximum probability
        Returns
        -------------
        vit_track1 : array
           array of indicies for track in obs1
        vit_track2 : array
           array of indicies for track in obs2
        vit_track: array
           array of indicies for track in the "reference" detector
        max_end_prob: float
           value of maxmimum end statistic
        """
        cdef long previous = 1
        cdef long[:] vit_trackr = np.zeros(self.length).astype(int)
        cdef long[:] vit_track2 = np.zeros(self.length).astype(int)
        cdef long[:] vit_track1 = np.zeros(self.length).astype(int)
        
        cdef float max_prob = np.array([self.V[self.length-1][i] for i in range(self.width)]).max()
        cdef long max_prob_index = np.array([self.V[self.length-1][i] for i in range(self.width)]).argmax()
        
        vit_trackr[self.length-1] = max_prob_index
        vit_track1[self.length-1] = self.det1[self.length-1][max_prob_index]
        vit_track2[self.length-1] = self.det2[self.length-1][max_prob_index]
        previous = self.prev[self.length-1][max_prob_index]
        

        for t in range(self.length-2,-1,-1):
            vit_trackr[t] = previous # insert previous step
            vit_track1[t] = self.det1[t][previous]
            vit_track2[t] = self.det2[t][previous]
            previous = self.prev[t][previous]
    
        self.vit_track1 = np.array(vit_track1)
        self.vit_track2 = np.array(vit_track2)
        self.vit_track = np.array(vit_trackr)
        self.max_end_prob = max_prob
        
        #return np.array(vit_track1), np.array(vit_track2), np.array(vit_trackr), np.array(val),np.array(prev),max_prob

        
class three_detector(object):
 
    def __init__(self, tr, obs1, obs2, obs3, prog = False):
        '''
        initialising viterbi class
        '''
        self.prog = prog
        self.tracks = self.run(tr, obs1, obs2, obs3)

    

    #---------------------------------------------------------
    # Three detector viterbi
    #--------------------------------------------------------

    def run(self,tr1, obs11, obs21, obs31):
        """
        viterbi algorithm for multiple detectors 2 at the moment
        Args
        ---------
        tr1: array
            transition matrix
        obs11: array
            observation from detector 1
        obs21: array
            observation from detector 2
        returns
        -----------
        vit_track1: array
            path in det 1
        vit_track2: array
            path in 2
        vit_trackr: array
            path in reference detector
        val: array
            values in reference detector
        prev: array
            previous positions
        det1: array
            det1 data
        det2: array
            det 2 data
        """
        #tr_len = np.arange(len(tr))-len(tr)/2
        #range_tr = np.arange(3)
        
    
        cdef double[:, :, :, :] tr = tr1
        cdef double[:, :] obs1 = obs11
        cdef double[:, :] obs2 = obs21
        cdef double[:, :] obs3 = obs31

        shape = np.shape(obs1)
        
        cdef int length = min(len(obs1),len(obs2),len(obs3))
        cdef int width = len(obs1[0])
        
        cdef double[:, :] val = np.ones(shape)*-1e6
        cdef int[:, :] prev = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det1 = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det2 = np.zeros(shape,dtype = np.int32)
        cdef int[:, :] det3 = np.zeros(shape,dtype = np.int32)
        
        cdef int tr_len = len(tr)
        
        cdef int i
        cdef int j
        cdef int k
        cdef int m
        cdef int t
        cdef int x
        
        cdef long max_pos
        cdef long[:] indicies
        cdef double max_val
        cdef double value
        cdef double temp
        #cdef double[:] max_vals
    
        for i in range(width):
            
            temp = -1e6
            for k in range(tr_len): 
                if i+k-1>=0 and i+k-1<width:
                    for m in range(tr_len):
                        if i+m-1>=0 and i+m-1<width:
                            for l in range(tr_len):
                                if i+l-1>=0 and i+l-1<width:
                                    value = obs1[0][i+k-1] + obs2[0][i+m-1] + obs3[0][i+l-1] + tr[1][k][m][l]
                                if value > temp:
                                    temp = value
                                    val[0][i] = temp
                                    prev[0][i] =  i
                                    det1[0][i] =  i+k-1 
                                    det2[0][i] =  i+m-1
                                    det3[0][i] =  i+l-1
                                else:
                                    continue

    
        # i is current frequency position
        # j is transition in reference detector
        # k is separation from other detector
        
        pbar = 0
        for t in range(1,length):
            pt = t/(float(length))*100.
            for i in range(width):
                
                temp = -1e6
                for j in range(tr_len):
                    if i+j-1>=0 and i+j-1<width:
                       for k in range(tr_len):
                            if i+k-1>=0 and i+k-1<width:
                                for m in range(tr_len):
                                     if i+m-1>=0 and i+m-1<width:
                                         for l in range(tr_len):
                                             if i+l-1>=0 and i+l-1<width:
                                                 value = obs1[t][i+k-1] + obs2[t][i+m-1] + obs3[t][i+l-1] + tr[j][k][m][l] + val[t-1][i+j-1]
                                                 if value > temp:
                                                     temp = value
                                                     val[t][i] = temp
                                                     prev[t][i] =  i+j-1
                                                     det1[t][i] =  i+k-1 
                                                     det2[t][i] =  i+m-1
                                                     det3[0][i] =  i+l-1
                                                 else:
                                                     continue
            
            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./length

        
    
        cdef long previous = 1
        cdef long[:] vit_trackr = np.zeros(length).astype(int)
        cdef long[:] vit_track1 = np.zeros(length).astype(int)
        cdef long[:] vit_track2 = np.zeros(length).astype(int)
        cdef long[:] vit_track3 = np.zeros(length).astype(int)
        
        cdef float max_prob = np.array([val[length-1][i] for i in range(width)]).max()
        cdef long max_prob_index = np.array([val[length-1][i] for i in range(width)]).argmax()
        
        vit_trackr[length-1] = max_prob_index
        vit_track1[length-1] = det1[length-1][max_prob_index]
        vit_track2[length-1] = det2[length-1][max_prob_index]
        vit_track3[length-1] = det3[length-1][max_prob_index]
        previous = prev[length-1][max_prob_index]
        

        for t in range(length-2,-1,-1):
            vit_trackr[t] = previous # insert previous step
            vit_track1[t] = det1[t][previous]
            vit_track2[t] = det2[t][previous]
            vit_track3[t] = det3[t][previous]
            previous = prev[t][previous]
    
        self.vit_track1    = np.array(vit_track1)
        self.vit_track2    = np.array(vit_track2)
        self.vit_track3    = np.array(vit_track3)
        self.vit_track = np.array(vit_trackr)
        self.V            = np.array(val)
        self.prev         = np.array(prev)
        self.max_end_prob = max_prob
    
        #return np.array(vit_track1), np.array(vit_track2), np.array(vit_track3), np.array(vit_trackr), np.array(val),np.array(prev), max_prob

class single_detector_mem_n(object):
 
    def __init__(self, tr, obs, prog = False):
        '''
        
        Find single detector algorithm with a memory, uses the summed pixel power as statistic

        initialising viterbi class
        '''
        self.prog = prog
        self.tracks = self.run(tr, obs)

    
    #--------------------------------------------------------------------------
    # n dimensional memory viterbi
    #------------------------------------------------------------------------

    def run(self, tr, obs):
        ''' Run viterbi algroithm with memory of n
        the amount of memory is determined by the transition matrix
        args
        -------
        tr : array
            transition matrix, any dimension greater that 1 - type: array
        obs: array
            observation - type: array
        
        returns
        -------
        vit_track: array
            optimum path through data
        val: array
            values of viterbi
        prev: array
            previous positions of viterbi
        '''
        
        if len(np.shape(tr))<=1:
            raise Exception("ERROR: Please use a transition matrix with dimensions larger than one")
        
        #np.insert(obs,0,-1e6,axis=1)
        #np.insert(obs,len(obs[0]),-1e6,axis=1)

        #cdef double[:] tr = tr1
        #cdef double[:, :] obs = obs1
        
        range_tr = np.arange(len(tr))
        cdef int n = len(tr)
        shape = np.shape(tr)
        cdef int length = len(shape)
        
        dimensions = (len(obs),len(obs[0]))
        dimensions = list(dimensions)
        
        cdef int length_t = dimensions[0]
        cdef int width = dimensions[1]
        
        cdef int i
        cdef int t
        cdef int k
        cdef int c
        
        cdef double value
        cdef double temp
        
        for i in range(len(shape)-1):
            dimensions.append(n)
        
        dimensions = tuple(dimensions)

        val    = np.ones(dimensions)*-1e6
        prev   = np.ones(dimensions)
    
        range_j = list(itertools.product(range(n),repeat=length-1))
    
        for i in range(width):
            for j in range_j:
                val[0][i][j]    = obs[0][i]
                prev[0][i][j]   = np.nan
    
        cdef float pbar = 0
        for t in range(1,length_t):
            obst = obs[t]
            valt = val[t-1]
            pt = t/float(length_t)*100.
            for i in range(width): 
                for j in range_j:
                    cond = True
                    for c in range(1,len(j)+1):
                        b = i+np.sum(j[:c])-len(j[:c])
                        if b >= 0 and b<width: 
                            continue
                        else:
                            cond = False
                            break
                    if cond:
                        temp = -1e6
                        for k in range(n):
                            m = list(j[1:])
                            m.append(k)
                            if t<length:
                                for n1 in range(len(m)):
                                    if n1>=t-1:
                                        m[n1] = 1
                            m = tuple(m)
                            value = obst[i] + tr[j[0]][m] + valt[i+j[0]-1][m]
    
                        
                            if value>temp:
                                temp = value
                                val[t][i][j] = temp
                                prev[t][i][j] = k

            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./length_t

        cdef int previous   = 1
        previous_1 = None
        vit_track   = [0]

        #cdef double max_end_prob = max(val[length-1][i] for i in range(width))

        temp = -1e6
        for i in range(width):
            for j in range_j:
                value = val[-1][i][j]
                if value>temp:
                    temp = value
                    vit_track[0] = i
                    previous = i
                    previous_1 = j
                    previous_2 = int(prev[-1][i][j])
    
                
        for t in range(length_t-2,-1,-1):
            vit_track.append(previous+previous_1[0]-1)
            previous += previous_1[0]-1
            previous_1 = list(previous_1[1:])
            previous_1.append(int(previous_2))
            previous_1 = tuple(previous_1)
            previous_2 = prev[t][previous][previous_1]
        
        vit_track = vit_track[::-1]

        self.vit_track = np.array(vit_track)
        self.V = np.array(val)
        self.prev = np.array(prev)
        self.max_end_prob = None # needs to be fixed
        
        #return (vit_track,val,prev)

class two_detector_mem_n(object):
 
    def __init__(self, tr, obs1, obs2, prog = False):
        '''
        initialising viterbi class
        '''
        self.prog = prog
        self.tracks = self.run(tr, obs1, obs2)

    
    #------------------------------------------------------------------------
    # Multi viterbi vit n dimensional memory
    #-----------------------------------------------------------------------
    

    def run(self, tr, obs11, obs22):
        ''' Run viterbi algroithm with memory of n
        the amount of memory is determined by the transition matrix
        args
        -------
        tr : transition matrix, any dimension greater that 1 - type: array
        obs: observation - type: array
        
        returns
        -------
        vit_track: optimum path through data
        '''
        
        if len(np.shape(tr))<=1:
            raise Exception("ERROR: Please use a transition matrix with dimensions larger than one")
        
        range_tr = np.arange(len(tr))
        
        cdef double[:, :] obs1 = obs11
        cdef double[:, :] obs2 = obs22
        
        cdef int n      = len(tr)
        shape           = np.shape(tr)
        cdef int length = len(shape)
        
        dimensions = (min(len(obs1),len(obs2)),min(len(obs1[0]),len(obs2[0])))
        dimensions = list(dimensions)
        
        cdef int length_t = dimensions[0]
        cdef int width = dimensions[1]
        
        cdef int i
        cdef int k
        cdef int m
        cdef int j1
        cdef int n1
        cdef int b
        cdef int t
        cdef double value
        cdef double temp

        
        for i in range(len(shape)-3):
            dimensions.append(n)
        
        dimensions = tuple(dimensions)

        val    = np.ones(dimensions)*-1e6
        prev   = np.ones(dimensions)
        pos1   = np.ones(dimensions)
        pos2   = np.ones(dimensions)
        
    
        range_j = list(itertools.product(range(n),repeat=length-3))
    
        for i in range(width):
            for j in range_j:
                temp = -1e6
                for k in range(n):
                    if i+k-1>=0 and i+k-1<width:
                        for m in range(n):
                            if i+m-1>=0 and i+m-1<width:
                                value = obs1[0][i+k-1] + obs2[0][i+m-1] + tr[tuple(np.ones(len(shape)-2).astype('int'))][k][m]
                                if value>temp:
                                    temp = value
                                    val[0][i][j]    = value
                                    prev[0][i][j]   = np.nan
                                    pos1[0][i][j]   = k
                                    pos2[0][i][j]   = m
    
        pbar = 0
        for t in range(1,length_t):
            obs1t = obs1[t]
            obs2t = obs2[t]
            valt = val[t-1]
            pt = t/float(length_t)*100.
            for i in range(width): 
                for j in range_j:
                    cond = True
                    for c in range(1,len(j)+1):
                        b = i+np.sum(j[:c])-len(j[:c])
                        if b >= 0 and b<width: 
                            continue
                        else:
                            cond = False
                            break
                    if cond:
                        temp = -1e6
                        for j1 in range(n):
                            a = list(j[1:])
                            a.append(j1)
                            if t<length:
                                for n1 in range(len(a)):
                                    if n1>=t-1:
                                        a[n1] = 1
                            a = tuple(a)
                            for k in range(n):
                                if i+k-1>=0 and i+k-1<width:
                                    for m in range(n):
                                        if i+m-1>=0 and i+m-1<width:
                                            value = obs1t[i+k-1] + obs2t[i+m-1] + tr[j][j1][k][m] + valt[i+j[0]-1][a]
                                            if value > temp:
                                                temp = value
                                                val[t][i][j]  = temp
                                                prev[t][i][j] = j1
                                                pos1[t][i][j] = k
                                                pos2[t][i][j] = m
                        """
                        for k in range(length):
                            m = list(j[1:])
                            m.append(k)x
                            if t<length:
                                for n1 in range(len(m)):
                                    if n1>=t-1:
                                        m[n1] = 1
                            m = tuple(m)
                            value = obst[i] + tr[j[0]][m] + valt[i+j[0]-1][m]
                            if value>temp:
                                temp = value
                                val[t][i][j] = temp
                                prev[t][i][j] = k
                        """
                    
            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./length_t

        previous   = 1
        previous_1 = None
        vit_track1  = [0]
        vit_track2  = [0]
        vit_trackr  = [0]

        temp = -1e6
        for i in range(width):
            for j in range_j:
                value = val[-1][i][j]
                if value>temp:
                    temp = value
                    vit_trackr[0] = i
                    vit_track1[0] = i+pos1[-1][i][j]-1
                    vit_track2[0] = i+pos2[-1][i][j]-1
                    previous = i
                    previous_1 = j
                    previous_2 = int(prev[-1][i][j])
    
        
        for t in range(length_t-2,-1,-1):
            vit_trackr.append(previous+previous_1[0]-1)
            previous += previous_1[0]-1
            previous_1 = list(previous_1[1:])
            previous_1.append(int(previous_2))
            previous_1 = tuple(previous_1)
            vit_track1.append(previous+pos1[t][previous][previous_1]-1)
            vit_track2.append(previous+pos2[t][previous][previous_1]-1)
            previous_2 = prev[t][previous][previous_1]
        
        vit_trackr = vit_trackr[::-1]   
        vit_track1 = vit_track1[::-1] 
        vit_track2 = vit_track2[::-1]

        self.vit_track1 = np.array(vit_track1)
        self.vit_track2 = np.array(vit_track2)
        self.vit_track = np.array(vit_trackr)
        self.V = np.array(val)
        self.prev = np.array(prev)
        
        #return (vit_track1,vit_track2,vit_trackr,val,prev)


### -----------------------------------------------------------------

## Single detector algorithm where data can include large gaps 

### ----------------------------------------------------------------

class single_detector_gaps(object):

    def __init__(self, tr, obs,times,tsft):
        """
        single detector viterbi algorithm where the data can include large gaps in data
        NOTE: This is for large gaps in data, the single_detector algorithm works faster if gaps are filled with the mean of data

        Parameters
        -----------------
        tr: list
            transition matrix elements
        obs: array
            an array of the sfts
        times: list
            times corresponding to each of the sfts
        tsft: float
            the length of each sfts in s
        Returns
        -------------------
        self.vit_track: list
            optimum path through dataset
        self.vit_track_times: list
            list of the times corresponding to each of the path elements
        """
        self.vits, self.sts, self.gps = self.vit_gaps(obs, times, tr=tr, tsft=tsft)
        self.vit_track, self.vit_track_times = self.get_track(self.vits, self.sts, self.gps,tsft=tsft)
        self.max_end_prob = max(self.vits[-1].V[-1])
        
    def max_track(self,tr,length,diff,bnd,ind,band_width):
        tr_val = np.arange(len(tr)).astype(int) - np.floor(len(tr)/2).astype(int)
        result = [seq for seq in itertools.combinations_with_replacement(tr_val, length) if sum(seq) == diff]
        max_seq = None
        thresh = -np.inf
        thresh2 = -np.inf
        #print ind,bnd, diff, [np.cumsum(s)+ind for s in result]
        for seq in result:
            #print seq
            abspath = np.array(np.cumsum(seq) + ind)
            if np.any(abspath < 0) == True or np.any(abspath >= band_width) == True:
                continue
            else:
                s = sum([tr[elem+1] for elem in seq]) 
                l = len(np.where(np.array(seq) == 0)[0])
                if s > thresh:
                    max_seq = np.insert(abspath,0,ind)[:-1] 
                    thresh = s
                elif s == thresh and l > thresh2:
                    max_seq = np.insert(abspath,0,ind)[:-1] 
                    thresh = s
                    thresh2 = l

        return max_seq,s

    def gap_run(self,viterbi, band_width, diff, tsft, tr):
        length = diff - 1
        max_vals = np.zeros(band_width)
        max_paths = np.zeros((band_width,length))
        #print band_width, length
        for bnd in range(band_width):
            low = bnd - length
            high = bnd + length + 1
            if low < 0:
                low = 0
            if high >= band_width:
                high = band_width
            thres = 0
            vend = viterbi.V[-1,low:high]
        
            val = 0
            ind = 0
            for j in range(len(vend)):
                if vend[j] > thres:
                    val = vend[j]
                    ind = low + j
                    thres = vend[j] 
            diff_ind = bnd-ind
            gaptrack = self.max_track(tr,length,diff_ind,bnd,ind,band_width)
            #print bnd,ind,low,high,diff_ind,length,gaptrack
        
            max_paths[bnd] = np.array(gaptrack[0])
            max_vals[bnd] = gaptrack[1] + val + length*2
    
        return max_vals, max_paths

    def find_path(self,prev,end_ind):
        length = len(prev)
        previous = prev[-1][end_ind]
        vit_track =  np.zeros(length,dtype=np.int32)
        vit_track[-1] = end_ind
        for t in range(length-2,-1,-1):
            vit_track[t] = previous # insert previous step
            previous = prev[t][previous]
        
        return vit_track
        

    def vit_gaps(self,data1,epochs1,tr,tsft):

        band_width = len(data1[0])
        vit = []
        starts = []
        gappaths = []
        gapend = 0
        vitstart = epochs1[0]
        max_vals = None
    

        data = []
        epochs = []
        for i,idx in enumerate(epochs1):
            if i == 0 or i == len(epochs1)-1:
                data.append(data1[i])
                epochs.append(idx)
            elif i > 0 and i != len(epochs1)-1:
                diff = epochs1[i]-epochs1[i-1]
                if diff == 2:
                    data.append(np.ones(band_width)*2)
                    epochs.append(idx-tsft)
                    data.append(data1[i])
                    epochs.append(idx)
                else:
                    data.append(data1[i])
                    epochs.append(idx)
        data = np.array(data)
        epochs = np.array(epochs)
    
        for i,idx in enumerate(epochs):
            if i == 0 :
                pass
            elif i > 0:# and i != len(epochs)-1:
                diff = epochs[i]-epochs[i-1]
                diff_ind = int(diff/tsft)
                if diff_ind >= 2 and diff_ind < band_width + 1:
                    if max_vals is not None:
                        viterbi = single_detector(tr, np.vstack([max_vals,data[gapend:i]]))
                        starts.append(np.arange(len(viterbi.vit_track))*tsft + vitstart - tsft)
                        vitstart = idx 
                    else:
                        viterbi = single_detector(tr, data[gapend:i])
                        starts.append(np.arange(len(viterbi.vit_track))*tsft + vitstart)
                        vitstart = idx 
        
                    vit.append(viterbi)
                    gapend = i
                    max_vals, max_paths = self.gap_run(viterbi, band_width, diff_ind, tsft, tr)
                    #print max_vals, max_paths
                    gappaths.append(max_paths)
                    if i == len(epochs)-1:
                        viterbi = single_detector(tr, np.vstack([max_vals,data[gapend:]]))
                        starts.append(np.arange(len(viterbi.vit_track))*tsft + idx - tsft)
                        vit.append(viterbi)
            
                elif diff_ind >= band_width + 1:
                
                    diff1 = band_width + 1
                    if max_vals is not None:
                        viterbi = single_detector(tr, np.vstack([max_vals,data[gapend:i]]))
                        starts.append(np.arange(len(viterbi.vit_track))*tsft + vitstart - tsft)
                        vitstart = idx 
                    else:
                        viterbi = single_detector(tr, data[gapend:i])
                        starts.append(np.arange(len(viterbi.vit_track))*tsft + vitstart)
                        vitstart = idx
        

                    vit.append(viterbi)
                    gapend = i
                    max_vals, max_paths = self.gap_run(viterbi, band_width , diff1, tsft, tr)
                    #print max_vals, max_paths
                    gappaths.append(max_paths) 
                    if i == len(epochs)-1:
                        viterbi = single_detector(tr, np.vstack([max_vals,data[gapend:]]))
                        starts.append(np.arange(len(viterbi.vit_track))*tsft + idx - tsft)
                        vit.append(viterbi)
                  
                else:
                    continue
                
        return vit,starts, gappaths

    def get_track(self,vit,starts,gappaths,tsft=1):
        vit_track = []
        vit_track_time = []
        gpb = False
        gp = None
        if len(starts) > 1:
            for v in range(len(starts))[::-1]:
                if gpb:
                    p1 = self.find_path(vit[v].prev,int(vit_track[-1]))[::-1]
                    vit_track.extend(p1[1:])
                    vit_track_time.extend(starts[v][::-1][1:])
                else:
                    p1 = vit[v].vit_track[::-1]
                    vit_track.extend(p1)
                    vit_track_time.extend(starts[v][::-1])
                st = vit_track[-1]
                if v-1 >= 0:
                    gp = gappaths[v-1][st][::-1]
                    vit_track.extend(gp)
                    vit_track_time.extend(np.array(np.arange(len(gp))*tsft + starts[v-1][-1])[::-1])#+np.array(range(len(gp))[::-1]))
                    #vit_track_time.extend(range(len(gp))[::-1])
                    gpb = True
                else:
                    gpb = False
        else:
            vit_track = vit[0].vit_track
        return vit_track[::-1], vit_track_time[::-1]




'''

class N_detector(object):
 
    def __init__(self, tr, obs):
        """
        UNTESTED: DO NOT USE
        initialising viterbi class
        """
        self.prog = prog
        self.tracks = self.run(tr, obs)

    
    #------------------------------------------------------------------------
    # Viterbi with N detectors
    #-----------------------------------------------------------------------
    

    def run(self, tr, mobs):
        """

        Run viterbi algroithm with memory of n
        the amount of memory is determined by the transition matrix
        args
        -------
        tr : transition matrix, any dimension greater that 1 - type: array
        obs: observation - type: array
        
        returns
        -------
        vit_track: optimum path through data
        """
        
        if len(np.shape(tr))<=1:
            raise Exception("ERROR: Please use a transition matrix with dimensions larger than one")
        
        range_tr = np.arange(len(tr))
        
        n_detectors = len(mobs)

        cdef double[:, :, :] obs_list = mobs
        
        cdef int n      = len(tr)
        shape           = np.shape(tr)
        cdef int length = len(shape)
        
        min_len1 = np.inf
        min_len2 = np.inf
        for obs in obs_list:
            if len(obs) < min_len1:
                min_len1 = len(obs)
            if len(obs[0]) < min_len2:
                min_len2 = len(obs[0])
            
        dimensions = (min_len1, min_len2)
        dimensions = list(dimensions)
        
        cdef int length_t = dimensions[0]
        cdef int width = dimensions[1]
        
        cdef int i
        cdef int k
        cdef int m
        cdef int j1
        cdef int n1
        cdef int b
        cdef int t
        cdef double value
        cdef double temp

        
        for i in range(len(shape)-3):
            dimensions.append(n)
        
        dimensions = tuple(dimensions)

        val    = np.ones(dimensions)*-1e6
        prev   = np.ones(dimensions)
        pos1   = np.ones(dimensions)
        pos2   = np.ones(dimensions)
        
    
        range_j = list(itertools.product(range(n),repeat=n_detectors))
    
        for i in range(width):
            temp = -1e6
            for k in range(n):
                if i+k-1>=0 and i+k-1<width:
                    for m in range(n):
                        if i+m-1>=0 and i+m-1<width:
                            value = obs1[0][i+k-1] + obs2[0][i+m-1] + tr[tuple(np.ones(len(shape)-2).astype('int'))][k][m]
                            if value>temp:
                                temp = value
                                val[0][i]    = value
                                prev[0][i]   = np.nan
                                pos1[0][i]   = k
                                pos2[0][i]   = m
    
        pbar = 0
        for t in range(1,length_t):
            obs1t = obs1[t]
            obs2t = obs2[t]
            valt = val[t-1]
            pt = t/float(length_t)*100.
            for i in range(width): 
                for j in range_j:
                    cond = True
                    for c in range(1,len(j)+1):
                        b = i+np.sum(j[:c])-len(j[:c])
                        if b >= 0 and b<width: 
                            continue
                        else:
                            cond = False
                            break
                    if cond:
                        temp = -1e6
                        for j1 in range(n):
                            a = list(j[1:])
                            a.append(j1)
                            if t<length:
                                for n1 in range(len(a)):
                                    if n1>=t-1:
                                        a[n1] = 1
                            a = tuple(a)
                            for k in range(n):
                                if i+k-1>=0 and i+k-1<width:
                                    for m in range(n):
                                        if i+m-1>=0 and i+m-1<width:
                                            value = obs1t[i+k-1] + obs2t[i+m-1] + tr[j][j1][k][m] + valt[i+j[0]-1][a]
                                            if value > temp:
                                                temp = value
                                                val[t][i][j]  = temp
                                                prev[t][i][j] = j1
                                                pos1[t][i][j] = k
                                                pos2[t][i][j] = m
                        """
                        for k in range(length):
                            m = list(j[1:])
                            m.append(k)x
                            if t<length:
                                for n1 in range(len(m)):
                                    if n1>=t-1:
                                        m[n1] = 1
                            m = tuple(m)
                            value = obst[i] + tr[j[0]][m] + valt[i+j[0]-1][m]
                            if value>temp:
                                temp = value
                                val[t][i][j] = temp
                                prev[t][i][j] = k
                        """
                    
            if self.prog == True:
                if pt>pbar:
                    stdout.write('\r{} %'.format(round(pt)))
                    stdout.flush()
                    pbar+=100./length_t

        previous   = 1
        previous_1 = None
        vit_track1  = [0]
        vit_track2  = [0]
        vit_trackr  = [0]

        temp = -1e6
        for i in range(width):
            for j in range_j:
                value = val[-1][i][j]
                if value>temp:
                    temp = value
                    vit_trackr[0] = i
                    vit_track1[0] = i+pos1[-1][i][j]-1
                    vit_track2[0] = i+pos2[-1][i][j]-1
                    previous = i
                    previous_1 = j
                    previous_2 = int(prev[-1][i][j])
    
        
        for t in range(length_t-2,-1,-1):
            vit_trackr.append(previous+previous_1[0]-1)
            previous += previous_1[0]-1
            previous_1 = list(previous_1[1:])
            previous_1.append(int(previous_2))
            previous_1 = tuple(previous_1)
            vit_track1.append(previous+pos1[t][previous][previous_1]-1)
            vit_track2.append(previous+pos2[t][previous][previous_1]-1)
            previous_2 = prev[t][previous][previous_1]
        
        vit_trackr = vit_trackr[::-1]   
        vit_track1 = vit_track1[::-1] 
        vit_track2 = vit_track2[::-1]

        self.vit_track1 = np.array(vit_track1)
        self.vit_track2 = np.array(vit_track2)
        self.vit_track = np.array(vit_trackr)
        self.V = np.array(val)
        self.prev = np.array(prev)
        
        #return (vit_track1,vit_track2,vit_trackr,val,prev)

'''

    



#!/usr/bin/env python
from __future__ import print_function
import matplotlib
matplotlib.use("agg")
import soapcw as soap
from soapcw import cw
import sys
import os
import subprocess
import numpy as np
import matplotlib.pyplot as plt
import time
import argparse
import configparser
import pickle
import h5py
from astropy.time import Time
import sys
import fcntl
import torch
import lalpulsar
import lal
from .soap_config_parser import SOAPConfig
from collections import OrderedDict

def set_limit(data,cut=180):
    """
    set the limits for matplotlib plot whet an upper cut has been applied
    """
    if data.max() > cut:
        vmax = cut
    else:
        vmax = data.max()
    vmin = data.min()
    return vmin,vmax

def get_sfts_in_range(tmin,tmax,sftlist):
    """
    load the sfts within some tmin and tmax range
    """
    in_range = []
    
    for sft in sftlist:
        stt = float(sft.split("-")[-2])
        #print(stt - tmin,tmax-stt)
        if stt >= tmin and stt < tmax:
            in_range.append(sft)
    return in_range

def get_sft_files(sftpaths):
    """
    look through sftpaths to find all sft files
    """
    sftlist = []
    if type(sftpaths) == str:
        sftpaths = sftpaths.split(",")
    for sftpath in sftpaths:
        #print(sftpath)
        for path, subdirs, files in os.walk(sftpath):
            #print(path, subdirs, files)
            for name in files:
                if name != "" and name.endswith(".sft"):
                    sftlist.append(os.path.join(path, name))
    return sftlist

def get_sft_files_find(output_dir, sftpaths, overwrite_file=False):
    sftlist = []
    if type(sftpaths) == str:
        sftpaths = sftpaths.split(",")
    fnames = []
    for sftpath in sftpaths:
        pathsplit = sftpath.split('/')
        pathappend = pathsplit[-2] if pathsplit[-1] == "" else pathsplit[-1]
        fname = os.path.join(output_dir, f"{pathappend}_sftfile.txt")
        fnames.append(fname)
        sftpath = sftpath.strip()
        print("sftfilename", fname)
        if not os.path.isfile(fname) or overwrite_file:
            with open(fname, "w") as f:
                findout1 = subprocess.Popen([
                        "find",
                        sftpath,
                        "-name",
                        "*.sft"
                    ], 
                    stdout=subprocess.PIPE,
                    shell=False)
                findout = subprocess.run([
                    "sort"],
                    stdin=findout1.stdout,
                    stdout=f,
                    shell=False
                )

                print(findout)
        else:
            print("SFT filelist exists")


    return fnames

def av_noise(noise,nsft=48):
    """
    get the median noise floor for every nsft sfts
    """
    av_noise = []
    
    for i in np.arange(len(noise))[::nsft]:
        if i + nsft > len(noise) - 1:
            av_noise.append(np.nanmedian(noise[i:]))
        else:
            av_noise.append(np.nanmedian(noise[i:i+nsft]))
    
    return np.array(av_noise)


def run_soap_onedet(
    config,
    degfree, 
    sft, 
    ind_start, 
    ind_end
    ):

    tr = soap.tools.transition_matrix(config["transitionmatrix"]["left_right_prob"])

    # run soap search using the two detector line aware statistic (one detector in gaps)
    # SOAP using SNR based statistic
    soaprunH1 = soap.single_detector(tr,sft.H1.downsamp_summed_norm_sft_power[:,ind_start:ind_end])
    soaprunL1 = soap.single_detector(tr,sft.L1.downsamp_summed_norm_sft_power[:,ind_start:ind_end])

    return soaprunH1, soaprunL1

def load_sft_band(config, filenames, minfreq, maxfreq, tmin, tmax, verbose = False):

    print("summedtype", type(config["data"]["n_summed_sfts"]))
    sft = cw.LoadSFT(
        filenames,
        norm=True,
        filled=True,
        save_rngmed=True,
        remove_sft=True,
        summed=config["data"]["n_summed_sfts"],
        fmin=minfreq,
        fmax=maxfreq,
        tmin=tmin,
        tmax=tmax,
        verbose=True
        )

    # set how many frequency bins to average over for each frequency band (based on spread of signal in 1800s sft)
    # hardcoded at the moment 
    for i, (minband, maxband) in enumerate(zip(config["data"]["band_starts"], config["data"]["band_ends"])):
        if minband <= minfreq < maxband:
            stride = config["data"]["strides"][i]
            degfree = int(2 * config["data"]["n_summed_sfts"] * stride)
            plotcut = stride * 180
            width = config["data"]["band_widths"][i]
            break
        else:
            continue


    if verbose:
        print(f"Nsum: {config['data']['n_summed_sfts']}")
        print("H1shape: {}, nsft: {}".format(np.shape(sft.H1.norm_sft_power), sft.H1.nsft))
        print("L1shape: {}, nsft: {}".format(np.shape(sft.L1.norm_sft_power), sft.L1.nsft))

    sft.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)

    if "H1" not in sft.det_names:
        sft.H1 = cw.SFT()
        sft.H1.downsamp_summed_norm_sft_power = np.ones(sft.L1.downsamp_summed_norm_sft_power.shape)*np.nan
        sft.det_names = ["H1", "L1"]
        sft.H1.rng_med = sft.L1.rng_med * np.nan
    elif "L1" not in sft.det_names:
        sft.L1 = cw.SFT()
        sft.L1.downsamp_summed_norm_sft_power = np.ones(sft.H1.downsamp_summed_norm_sft_power.shape)*np.nan
        sft.det_names = ["H1", "L1"]
        sft.L1.rng_med = sft.H1.rng_med * np.nan


    # print some variables shapes to confirm what is expected
    if verbose:
        print("H1shape downsamp: {}, freq: {}".format(np.shape(sft.H1.downsamp_summed_norm_sft_power), minfreq))
        print("L1shape downsamp: {}, freq: {}".format(np.shape(sft.L1.downsamp_summed_norm_sft_power), minfreq))
        print("H1shape sum: {}, freq: {}".format(np.shape(sft.H1.summed_norm_sft_power), minfreq))
        print("L1shape sum: {}, freq: {}".format(np.shape(sft.L1.summed_norm_sft_power), minfreq))
        print("H1epochs len: {}, sumep: {}, start {} end {}".format(np.shape(sft.H1.epochs), np.shape(sft.H1.summed_epochs), sft.H1.summed_epochs[0], sft.H1.summed_epochs[-1]))
        print("L1epochs len: {}, sumep: {},  start {} end {}".format(np.shape(sft.L1.epochs), np.shape(sft.L1.summed_epochs), sft.L1.summed_epochs[0], sft.L1.summed_epochs[-1]))


    return sft, stride, degfree, plotcut, width

def plot_setup(
    fig, 
    ax, 
    plotcut, 
    H_temp_sft, 
    L_temp_sft, 
    temp_noise,
    sum_nsft, 
    soaprunH1,
    soaprunL1, 
    label_fontsize, 
    tick_fontsize, 
    xticks, 
    xticklabels, 
    xticktimes,
    yticks, 
    yticklabels):
    """_summary_

    Args:
        fig (_type_): _description_
        ax (_type_): _description_
        plotcut (_type_): _description_
        H_temp_sft (_type_): _description_
        L_temp_sft (_type_): _description_
        sum_nsft (_type_): _description_
        soaprun (_type_): _description_
        label_fontsize (_type_): _description_
        tick_fontsize (_type_): _description_
        xticks (_type_): _description_
        xticklabels (_type_): _description_
        yticks (_type_): _description_
        yticklabels (_type_): _description_

    Returns:
        _type_: _description_
    """

    # plot images 
    H_imlim = set_limit(H_temp_sft, cut=plotcut)
    H_im = ax[0].imshow(H_temp_sft.T,cmap="cividis",aspect="auto",origin="lower",vmin=H_imlim[0],vmax=H_imlim[1], interpolation="none")
    L_imlim = set_limit(L_temp_sft, cut=plotcut)
    L_im = ax[1].imshow(L_temp_sft.T,cmap="cividis",aspect="auto",origin="lower",vmin=L_imlim[0],vmax=L_imlim[1], interpolation="none")
    imvh1 = ax[2].imshow(soaprunH1.vitmap.T,cmap="cividis",aspect="auto",origin="lower", interpolation="none")
    imvl1 = ax[3].imshow(soaprunL1.vitmap.T,cmap="cividis",aspect="auto",origin="lower", interpolation="none")
            
    # plot track power s
    Hpwline, = ax[4].plot(np.arange(sum_nsft),np.array(soap.tools.track_power(soaprunH1.vit_track,H_temp_sft)),color="C3",label="H1")
    Lpwline, = ax[4].plot(np.arange(sum_nsft),np.array(soap.tools.track_power(soaprunL1.vit_track,L_temp_sft)),color="C2",label="L1")

    # plot viterbi statistic
    Hstatline, = ax[5].plot(np.arange(sum_nsft),np.array(soap.tools.stat_on_path(soaprunH1.vit_track,soaprunH1.V)),color="C3",label="H1")
    Lstatline, = ax[5].plot(np.arange(sum_nsft),np.array(soap.tools.stat_on_path(soaprunL1.vit_track,soaprunL1.V)),color="C2",label="L1")

    # plot noise 
    Hnoise, = ax[6].plot(np.arange(sum_nsft),temp_noise["H1"],color="C3",label="H1")
    Lnoise, = ax[6].plot(np.arange(sum_nsft),temp_noise["L1"],color="C2",label="L1")

                
    # set colorbar parameters
    Hcbar = fig.colorbar(H_im,ax = ax[0])
    Lcbar = fig.colorbar(L_im,ax = ax[1])
    cbarvh1 = fig.colorbar(imvh1,ax = ax[2])
    cbarvl1 = fig.colorbar(imvl1,ax = ax[3])
    Hcbar.set_label("Normalised SFT \n power (H1)",fontsize = label_fontsize)
    Lcbar.set_label("Normalised SFT \n power (L1)",fontsize = label_fontsize)
    cbarvh1.set_label("Normalised \n log-odds (H1)",fontsize = label_fontsize)
    cbarvl1.set_label("Normalised \n log-odds (L1)",fontsize = label_fontsize)
    Hcbar.ax.tick_params(labelsize=tick_fontsize)
    Lcbar.ax.tick_params(labelsize=tick_fontsize)
    cbarvh1.ax.tick_params(labelsize=tick_fontsize)
    cbarvl1.ax.tick_params(labelsize=tick_fontsize)

    # set tick labels
    for i in range(6):
        ax[i].set_xticks(xticks)

    ax[5].set_xticklabels(xticklabels)

    for i in range(3):
        ax[i].set_yticks(yticks)
        ax[i].set_yticklabels(yticklabels)

    ax[5].xaxis.set_tick_params(rotation=0,labelsize=tick_fontsize)
    for i in range(6):
        ax[i].yaxis.set_tick_params(rotation=0,labelsize=tick_fontsize)

    # set axis labels
    ax[5].set_xlabel("Time - {}[s]".format(xticktimes[0].year),fontsize=label_fontsize)
    ax[0].set_ylabel("Frequency [Hz]",fontsize=label_fontsize)
    ax[1].set_ylabel("Frequency [Hz]",fontsize=label_fontsize)
    ax[2].set_ylabel("Frequency [Hz]",fontsize=label_fontsize)
    ax[3].set_ylabel("Frequency [Hz]",fontsize=label_fontsize)
    ax[4].set_ylabel("Normalised SFT power \n along track",fontsize=label_fontsize)
    ax[5].set_ylabel("Statistic along path",fontsize=label_fontsize)
    ax[6].set_ylabel("Mean noise floor \n for band",fontsize=label_fontsize)

    # set axis limits
    for i in [0,1,2,4,5]:
        ax[i].set_xlim([0,sum_nsft])

    ax[0].set_ylim([0,len(H_temp_sft[0])])
    ax[1].set_ylim([0,len(L_temp_sft[0])])
    ax[2].set_ylim([0,len(L_temp_sft[0])])
    ax[3].set_ylim([0,len(L_temp_sft[0])])
            
    ax[4].legend()
    ax[5].legend()

    for i in [4,5,6]:
        ax[i].grid()

    Hcbar.outline.set_visible(False)
    Lcbar.outline.set_visible(False)
    cbarvh1.outline.set_visible(False)
    cbarvl1.outline.set_visible(False)
            
    # set widths oof plots
    fig.subplots_adjust(hspace=0)

    box = ax[4].get_position()
    ax[4].set_position([box.x0, box.y0, box.width * 0.8 , box.height])

    box = ax[5].get_position()
    ax[5].set_position([box.x0, box.y0, box.width * 0.8 , box.height])
            
    box = ax[6].get_position()
    ax[6].set_position([box.x0, box.y0, box.width * 0.8 , box.height])

    return H_im, L_im, imvh1, imvl1, Hnoise, Lnoise, Hpwline, Lpwline, Hstatline

def update_plot(fig, ax, H_im, L_im, imvh1, imvl1, Hnoise, Lnoise, Hpwline, Lpwline, Hstatline, H_temp_sft, L_temp_sft, soaprunH1, soaprunL1, temp_noise, yticks, yticklabels, plotcut, ):

        # set the yaxis labels for the first two figures, i.e. the sfts and feather plot
    for axs in fig.axes[:3]:
        try:
            axs.set_yticks(yticks)
            axs.set_yticklabels(yticklabels)
        except:
            pass
            
    # set the new data and limits for the sft plot
    H_im.set_data(H_temp_sft.T)
    L_im.set_data(L_temp_sft.T)
    H_imlim = set_limit(H_temp_sft, cut = plotcut)
    H_im.set_clim(vmin=H_imlim[0],vmax=H_imlim[1])
    L_imlim = set_limit(L_temp_sft, cut = plotcut)
    L_im.set_clim(vmin=L_imlim[0],vmax=L_imlim[1])
    
    # set the new data and limits for the feather plot
    # take the log of vitmap to make it more visible (also add small value so no taking log of 0)
    # vitmaps are normalised between 0 and 1
    imvh1.set_data(soaprunH1.vitmap.T)
    #imvh1.set_clim(vmin=soaprunH1.vitmap.min(),vmax=soaprunH1.vitmap.max())
    imvl1.set_data(soaprunL1.vitmap.T)
    #imvl1.set_clim(vmin=soaprunL1.vitmap.min(),vmax=soaprunL1.vitmap.max())
    
    # set the new data and limits for the noise power plot
    Hnoise.set_ydata(temp_noise["H1"])
    Lnoise.set_ydata(temp_noise["L1"])
    ax[6].relim()
    ax[6].autoscale_view()
    
    # set the new data and limits for the track power plot
    Hpwline.set_ydata(soap.tools.track_power(soaprunH1.vit_track,H_temp_sft))
    Lpwline.set_ydata(soap.tools.track_power(soaprunL1.vit_track,L_temp_sft))
    ax[4].relim()
    ax[4].autoscale_view()

    # set the new data and limits for the vitstat plot
    Hstatline.set_ydata(soap.tools.stat_on_path(soaprunH1.vit_track,soaprunH1.V))
    Hstatline.set_ydata(soap.tools.stat_on_path(soaprunL1.vit_track,soaprunL1.V))
    ax[5].relim()
    ax[5].autoscale_view()
    
    # refresh the figure
    fig.canvas.draw()
    fig.canvas.flush_events()
    
    # remove the soaprun track from the sft plot
    """
    if num !=0:
        try:
            Hline.remove()
            Lline.remove()
        except:
            pass
    """

def write_locked_file(filepath, data, write_type="rb"):
    # write table of statistics 
    with open(filepath,write_type) as f:  
        while True:
            try:
                fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
                total_data = pickle.load(f) 
                total_data.update(data)
                fcntl.flock(f, fcntl.LOCK_UN)
                break
            except IOError or EOFError:
                time.sleep(0.1)
    return total_data



def run_soap_in_band(config, minfreq, maxfreq, weekstarttime=None, verbose = False, force_overwrite = False):
    """_summary_

    Args:
        config (_type_): _description_
        sftfiles (_type_): _description_
        minfreq (_type_): _description_
        maxfreq (_type_): _description_
        verbose (bool, optional): _description_. Defaults to False.

    Raises:
        Exception: _description_
        Exception: _description_
        Exception: _description_
        Exception: _description_
    """
    
    # get the sft files and start/end times
    outpath = os.path.join(*[config["output"]["save_directory"],config["data"]["obs_run"]])    

    if not os.path.isdir(outpath):
        os.makedirs(outpath)

    if weekstarttime is None:
        outpath_save = os.path.join(outpath,config["output"]["sub_directory"])
        if os.path.isdir(outpath_save):
            print(f"This week has already been generated: {outpath_save}")
            if force_overwrite == False:
                return 0
            else:
                print(f"WARNING!!! - Overwriting current existing directory: {outpath_save}")

    # get the list of sft filenames by iterating over subdirs 
    # this saves the filelise to a file
    sftfiles = config["input"]["load_directory"]
    print("num files: ", len(sftfiles))
    start_load = time.time()
    sft_filelists = get_sft_files_find(outpath, sftfiles, overwrite_file=config["output"]["overwrite_files"])

    print("SFT filelist", sft_filelists)
    # load in the sft filelists 
    sftlist = []
    for sftfile in sft_filelists:
        with open(sftfile,"r") as f:
            sftlist.extend(f.readlines())
    
    if len(sftlist) == 0:
        raise Exception("No SFTs found")

    # get all the start times from the files
    try:
        sttime = [float(os.path.basename(name).split("-")[-2]) for name in sftlist]
    except Exception as e:
        print(os.path.basename(name))
        raise Exception(e)

    # define start and end times
    print("numsfts: {}".format(len(sftlist)))
    print("Time find fnames: ", time.time() - start_load)
    if "start_time" in config["data"].keys():
        tmin = float(config["data"]["start_time"])
        tmax = float(config["data"]["end_time"])
    else:
        if weekstarttime is not None:
            weeklength = 3600*24*7
            tmin = weekstarttime
            tmax = tmin + weeklength
        else:
            tmin = np.min(sttime) 
            tmax = np.max(sttime) + 1800 # start of last sft plus length (in this case we are only using 1800s sfts)
            #tmax = min(sttime) + 48*10*1800 # smaller range for testing
    
    # get the sfts in the specified time range 
    start_range = time.time()
    sftlist = get_sfts_in_range(tmin,tmax,sftlist)
    time_getsftsrange = time.time() - start_range
    print("time_getsftsrange: ", time_getsftsrange)

    # write time of last run to file
    lastruntime = os.path.join(outpath, "lastruntime.txt")
    if os.path.isfile(lastruntime):
        with open(lastruntime,"r") as f:
            mnt = np.loadtxt(f)
    
    # initialise figure and axes
    fig, axes = None, None
    
    if len(sftlist) == 0:
        raise Exception("No SFTs found in range")

    # define start times in gps and date format
    start_time_gps = Time(tmin,format="gps")
    end_time_gps = Time(tmax,format="gps")
    startdatestring = start_time_gps.iso.split(" ")[0].replace("-","")
    enddatestring = end_time_gps.iso.split(" ")[0].replace("-","")

    # set output directories
    outpath_save = os.path.join(outpath,config["output"]["sub_directory"])

    if os.path.isdir(outpath_save):
        pathexists = True
    else:
        pathexists = False

    # combine all filenames infor format input to lalpulsar loading (via LoadSFT)
    filenames = ";".join([s.strip() for s in sftlist])

    # load in the sfts, normalise them to the running median and fill the gaps
    start_load_sft = time.time()
    sft, stride, degfree, plotcut, width = load_sft_band(config, filenames, minfreq, maxfreq, tmin, tmax, verbose=verbose)
    # print the time takes to load the data
    load_time = time.time() - start_load_sft
    print("Load time: ", load_time)

    # define varialbes of sft
    detector = sft.det_names[0]
    tsft = getattr(sft,detector).tsft
    sum_nsft = len(getattr(sft,detector).summed_epochs)
    bin_width = stride/float(tsft)
    nbins = len(getattr(sft,detector).downsamp_frequencies)
            
    # set the start time of data and the indicies of the frequency bands the sft will be split into
    tstart = getattr(sft,detector).epochs[0]
    #start_indicies = np.arange(minfreq*sft.tsft,maxfreq*sft.tsft,width*sft.tsft/2).astype(int)
    start_indicies = np.round(np.arange(0,nbins,width/(2*bin_width))).astype(int)[:-1]

    # arrays are for timing different parts of the run
    av_soapruns = []
    av_cnnrun = []
    av_images = []
    av_save = []


    # initialise the plot figure and set some parameters
    fig, ax = plt.subplots(nrows=6,figsize=(18,25),sharex=True)
    tick_fontsize = 13
    label_fontsize = 20
    
    table_data = OrderedDict()
    
    start_run = time.time()
    breakloop = False
    # loop over all of the sub-bands 


    # write time of last run to file
    lastruntime = os.path.join(outpath, "lastruntime.txt")
    if os.path.isfile(lastruntime):
        with open(lastruntime,"r") as f:
            mnt = np.loadtxt(f)
    
    # initialise figure and axes
    fig, axes = None, None
    
    if len(sftlist) == 0:
        raise Exception("No SFTs found")

    # define start times in gps and date format
    start_time_gps = Time(tmin,format="gps")
    end_time_gps = Time(tmax,format="gps")
    startdatestring = start_time_gps.iso.split(" ")[0].replace("-","")
    enddatestring = end_time_gps.iso.split(" ")[0].replace("-","")

    # set output directories
    if weekstarttime is not None:
        outpath_save = os.path.join(outpath,config["output"]["sub_directory"], f"week_{startdatestring}_{enddatestring}")
    else:
        outpath_save = os.path.join(outpath,config["output"]["sub_directory"])

    if os.path.isdir(outpath_save):
        pathexists = True
        print(f"This week has already been generated: {outpath_save}")
        if force_overwrite == False:
            return 0
        else:
            print(f"WARNING!!! - Overwriting current existing directory: {outpath_save}")
    else:
        pathexists = False
                
    outpath_save_plots = os.path.join(outpath_save,"plots")
    outpath_save_track = os.path.join(outpath_save,"tracks")
    output_save_files = os.path.join(outpath_save, "data")
            
    for direc in [outpath_save_plots,outpath_save_track, output_save_files]:
        if not os.path.isdir(direc):
            os.makedirs(direc)
        else:
            pass

    tablefile = os.path.join(output_save_files,f"table_{minfreq}_{maxfreq}.hdf5")
    if os.path.isfile(tablefile):
        if not force_overwrite:
            raise FileExistsError(f"File {tablefile} already exists, add force-overwrite to arguments to overwrite file" )

    # combine all filenames infor format input to lalpulsar loading (via LoadSFT)
    filenames = ";".join([s.strip() for s in sftlist])

    # load in the sfts, normalise them to the running median and fill the gaps
    start_load_sft = time.time()
    sft, stride, degfree, plotcut, width = load_sft_band(config, filenames, minfreq, maxfreq, tmin, tmax, verbose=verbose)
    # print the time takes to load the data
    load_time = time.time() - start_load_sft
    print("Load time: ", load_time)

    # define varialbes of sft
    detector = sft.det_names[0]
    tsft = getattr(sft,detector).tsft
    sum_nsft = len(getattr(sft,detector).summed_epochs)
    bin_width = stride/float(tsft)
    nbins = len(getattr(sft,detector).downsamp_frequencies)

            
    # set the start time of data and the indicies of the frequency bands the sft will be split into
    tstart = getattr(sft,detector).epochs[0]
    #start_indicies = np.arange(minfreq*sft.tsft,maxfreq*sft.tsft,width*sft.tsft/2).astype(int)

    ####################
    ### Key difference for line search is no overlapping bands
    ###################
    start_indicies = np.round(np.arange(0,nbins,width/(bin_width))).astype(int)[:-1]
    #####################

    
    # arrays are for timing different parts of the run
    av_soapruns = []
    av_cnnrun = []
    av_images = []
    av_save = []


    # initialise the plot figure and set some parameters
    fig, ax = plt.subplots(nrows=7,figsize=(18,25),sharex=True)
    tick_fontsize = 13
    label_fontsize = 20
    
    table_data = OrderedDict()
    
    start_run = time.time()
    breakloop = False
    # loop over all of the sub-bands 

    for num,start in enumerate(start_indicies):
        
        # set the minimum freuency and maximum frequency of the sub-band
        fmin,fmax = minfreq + start*bin_width, minfreq + start*bin_width + width 

        # start and end index of sfts
        ind_start = int(start)
        ind_end = int(start + width/bin_width)

        if breakloop == True:
            break
        if fmax >= maxfreq:
            breakloop = True

        if ind_end > nbins:
            continue

        # get the subband from the sft
        temp_sft = {}
        temp_noise = {}
        for det in getattr(sft,"det_names"):
            # get noise floor of subband
            temp_n = np.sqrt(np.nanmedian(getattr(sft,det).rng_med[:,ind_start:ind_end],axis=1))   
            print("tempn", det, np.shape(temp_n)) 
            # get nsfts and the average noise floor
            temp_noise[det] = av_noise(temp_n, nsft=config["data"]["n_summed_sfts"])[:sum_nsft]

            print("tempnoise", np.shape(temp_noise[det]))


        # run the single detector viterbi search and append the time taken to run to previous list
        st_vt = time.time()
        soaprunH1, soaprunL1 = run_soap_onedet(config, degfree, sft, ind_start, ind_end)
        av_soapruns.append(time.time()-st_vt)

        save_path = os.path.join(outpath_save_plots,"notrack_F{}_{}.png".format(np.round(fmin,2),np.round(fmax,2)))
        save_path_track = os.path.join(outpath_save_plots,"track_F{}_{}.png".format(np.round(fmin,2),np.round(fmax,2)))
        
        # Set up the table data
        table_save_data = OrderedDict()
        table_save_data["fmin"] = fmin
        table_save_data["fmax"] = fmax
        table_save_data["H1_viterbistat"] = soaprunH1.max_end_prob
        table_save_data["L1_viterbistat"] = soaprunL1.max_end_prob
        
        # if cnn model provided run the cnn on viterbi map  
        st_cnn = time.time()    

        table_save_data["plot_path"] = save_path_track


        for key, val in table_save_data.items():
            table_data.setdefault(key, [])
            if type(val) == str:
                val = val.encode("ascii")
            table_data[key].append(val)

        av_cnnrun.append(time.time()-st_cnn)

        # only for two detectors, will generalise laterc
        tracks = {}
        tracks["H1"]  = np.array(soaprunH1.vit_track)*bin_width + fmin
        tracks["L1"]  = np.array(soaprunL1.vit_track)*bin_width + fmin


        # save the tracks for each detector
        track_times = sft.H1.summed_epochs 
        with open(os.path.join(outpath_save_track,"track_F{}_{}.txt".format(np.round(fmin,2),np.round(fmax,2))),"w") as f:
            np.savetxt(f,np.transpose([track_times,tracks["H1"],tracks["L1"]]).astype(float))

        #table_data["{}_{}".format(fmin,fmax)] = table_save_data


        # set the save paths for the plots
        st_im = time.time()
                
        # create masked array to plots the sfts
        H_temp_sft = np.ma.masked_where(sft.H1.downsamp_summed_norm_sft_power[:,ind_start:ind_end] == degfree,sft.H1.downsamp_summed_norm_sft_power[:,ind_start:ind_end],copy=True)
        L_temp_sft = np.ma.masked_where(sft.L1.downsamp_summed_norm_sft_power[:,ind_start:ind_end] == degfree,sft.L1.downsamp_summed_norm_sft_power[:,ind_start:ind_end],copy=True)
                
        # define the signal track
        if len(sft.H1.downsamp_summed_norm_sft_power[:,ind_start:ind_end]) != len(soaprunH1.vit_track):
            raise Exception("length of data {} different to viterbi track {} , {}, {}, {}, {}".format(len(sft.H1.downsamp_summed_norm_sft_power[:,ind_start:ind_end]),len(soaprunH1.vit_track1) ,fmin,fmax,minfreq,maxfreq))
            
        # set gaps in viterbi track to nan for plots
        H_track = soaprunH1.vit_track
        H_track[sft.H1.downsamp_summed_norm_sft_power[:,0] == degfree] == np.nan
        L_track = soaprunL1.vit_track
        L_track[sft.L1.downsamp_summed_norm_sft_power[:,0] == degfree] == np.nan
        
        # set gaps to nan in noise estimates
        print("datashape", np.shape(sft.H1.downsamp_summed_norm_sft_power))
        for det in getattr(sft,"det_names"):
            temp_noise[det][sft.H1.downsamp_summed_norm_sft_power[:,0] == degfree] = np.nan
            
        if len(H_track) != sum_nsft:
            sum_nsft = len(H_track)
            print("sum_nsft not equal to length of viterbi")

        if len(H_track) != len(H_temp_sft):
            raise Exception("Viterbi track and tsft not same length")
                
        # set the xaxis of the plot to each month of the year and its date
        months = ["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]
        xticks = np.linspace(0,(sum_nsft-1),10)
        xticktimes = [Time(tstart + i*48*tsft,format='gps').datetime for i in xticks]
        xticklabels = ["{}-{}".format(tm.day,months[tm.month-1]) for tm in xticktimes]
        
        # set the yaxis of the plot as the frequency band
        yticks = np.linspace(0,len(H_temp_sft[0]),8).astype(int)
        yticklabels = np.round(np.linspace(fmin,fmax,8),2)
        
        # if its the first plot then set up labels and plots
        if num == 0:

            H_im, L_im, imvh1, imvl1, Hnoise, Lnoise, Hpwline, Lpwline, Hstatline = plot_setup(
                                                                                        fig, 
                                                                                        ax, 
                                                                                        plotcut, 
                                                                                        H_temp_sft, 
                                                                                        L_temp_sft, 
                                                                                        temp_noise,
                                                                                        sum_nsft, 
                                                                                        soaprunH1,
                                                                                        soaprunL1, 
                                                                                        label_fontsize, 
                                                                                        tick_fontsize, 
                                                                                        xticks, 
                                                                                        xticklabels, 
                                                                                        xticktimes,
                                                                                        yticks, 
                                                                                        yticklabels)
                    
        else:
            # for other plots just change data and axis labels (runs faster that generating new figure)
            
            update_plot(fig, ax, H_im, L_im, imvh1, imvl1, Hnoise, Lnoise, Hpwline, Lpwline, Hstatline, H_temp_sft, L_temp_sft, soaprunH1, soaprunL1, temp_noise, yticks, yticklabels, plotcut)

            Hline.set_data([], [])
            Lline.set_data([], [])

        # save the figure without the soaprun track on sft plot
        av_images.append(time.time() - st_im)
        st_sv = time.time()
        fig.savefig(save_path,bbox_inches="tight",dpi=60)
        
        if num == 0:
            # add the soaprun trck to sft plot
            Hline,  = ax[0].plot(np.arange(sum_nsft),H_track,color="red",label="Soaprun track",marker="o",ms=3)
            Lline, = ax[1].plot(np.arange(sum_nsft),L_track,color="red",label="Soaprun track",marker="o",ms=3)
        else:
            Hline.set_data(np.arange(sum_nsft),H_track)
            Lline.set_data(np.arange(sum_nsft),L_track)
        
        # save second figure with soaprun track
        fig.savefig(save_path_track,bbox_inches="tight",dpi=60)
        
        av_save.append(time.time() - st_sv)
        del H_temp_sft,L_temp_sft, temp_noise
     
    # write table of statistics 
    if os.path.isfile(tablefile):
        if not force_overwrite:
            new_tablefile = os.path.join(output_save_files,f"table_{minfreq}_{maxfreq}_{np.random.randint(low=0,high=1000)}.hdf5")
            print(f"File exists: {tablefile}, writing to {new_tablefile}")
            tablefile = new_tablefile
        else:
            tablefile = tablefile
        #with open(new_tablefile,"wb") as f:  
        #    pickle.dump(table_data, f)
    else:
        tablefile = tablefile

    with h5py.File(tablefile, "w", track_order=True) as hf:
        for key, val in table_data.items():
            try:
                hf.create_dataset(key, data=np.array(val))
            except TypeError as e:
                print("ERROR: ",key, val)
                raise TypeError(e)

        hf.attrs["Run Time"] =  np.array(time.time() - start_run)
        hf.attrs["Av SOAP run time"] = np.array(np.mean(av_soapruns))
        hf.attrs["Av image gen time"] =  np.array(np.mean(av_images))
        hf.attrs["Av im save time"] = np.array(np.mean(av_save))
        hf.attrs["Sum SOAP run time"] = np.array(np.sum(av_soapruns))
        hf.attrs["Sum image gen time"] =  np.array(np.sum(av_images))
        hf.attrs["Sum im save time"] = np.array(np.sum(av_save))
        hf.attrs["load SFT time"] = np.array(load_time)
        hf.attrs["fmin"] = fmin
        hf.attrs["fmax"] = fmax

        #with open(tablefile,"wb") as f:  
        #    pickle.dump(table_data, f)
            

    # show the times taken for each part of run
    print("Run time: ", time.time() - start_run)
    print("Av soaprun time: ", np.mean(av_soapruns))
    print("Av image generation time: ", np.mean(av_images))
    print("Av image save time: ", np.mean(av_save))
    print("Sum soaprun time: ", np.sum(av_soapruns))
    print("Sum image generation time: ", np.sum(av_images))
    print("Sum image save time: ", np.sum(av_save))





def read_data(filename):
    """
    Read the data for the cnn
    """
    with open(filename,"r") as f:
        # read image subtract offset and divide my standard deviation                                                                            
        image = pickle.load(f)
    image = image/image.max()

    global image_shape
    image_shape = image.shape
    reshape_image = image.reshape(image.shape[0],image.shape[1],1)
    return reshape_image

def run(model,test_data,device="cpu"):
    # set the evaluation mode                                                                                                                               
    model.eval()

    # test loss for the data                                                                                                                                
    test_loss = 0
    samples = {}
    # we don't need to track the gradients, since we are not updating the parameters during evaluation / testing                                            
    with torch.no_grad():
        test_batch, test_labels = torch.Tensor(np.array([val[0] for val in test_data])).to(device), torch.Tensor(np.array([val[1] for val in test_data])).to(device)

        outputs = model.test(test_batch)

    return outputs,test_labels.cpu().numpy()

def run_vitmap_cnn(modelfname, vit_data, size = (156,89)):
    """
    run the cnn
    """
    #model = load_model(modelfname)
    model = torch.load(modelfname, map_location="cpu").to("cpu")
    
    vit_resize = resize(vit_data, size, anti_aliasing=True)

    vit_resize = vit_resize/vit_resize.max()

    vit_resize = vit_resize.reshape(1,vit_resize.shape[0],vit_resize.shape[1])
    model.eval()
    with torch.no_grad():
        prob = model.test(torch.Tensor([vit_resize,]))

    return prob[0][0]

def run_vitmapstat_cnn(modelfname, vit_data, vit_stat, size = (156,89)):
    """
    run the cnn
    """
    #model = load_model(modelfname)
    model = torch.load(modelfname, map_location="cpu").to("cpu")
    
    vit_resize = resize(vit_data, size, anti_aliasing=True)

    vit_resize = vit_resize/vit_resize.max()

    vit_resize = vit_resize.reshape(1,vit_resize.shape[0],vit_resize.shape[1])
    model.eval()
    with torch.no_grad():
        prob = model.test(torch.Tensor([vit_resize,]), torch.Tensor([vit_stat,]))

    return prob[0][0]

def run_spect_cnn(modelfname, H_data, L_data, size = (156,89), degfree = 96):
    """
    run the cnn
    """
    #model = load_model(modelfname)
    model = torch.load(modelfname, map_location="cpu").to("cpu")
    
    H_resize = resize(H_data, size, anti_aliasing=True)
    L_resize = resize(L_data, size, anti_aliasing=True)

    H_resize = (H_resize - degfree)/(2*degfree)
    L_resize = (L_resize - degfree)/(2*degfree)

    H_resize = H_resize.reshape(1,H_resize.shape[0],H_resize.shape[1])
    L_resize = L_resize.reshape(1,L_resize.shape[0],L_resize.shape[1])
    spect_dat = np.vstack([H_resize, L_resize])

    model.eval()
    with torch.no_grad():
        prob = model.test(torch.Tensor([spect_dat,]))

    return prob[0][0]

def run_all_cnn(modelfname, vit_data, H_data, L_data, vit_stat, size = (156,89), degfree = 96):
    """
    run the cnn
    """
    #model = load_model(modelfname)
    model = torch.load(modelfname, map_location="cpu").to("cpu")

    vit_resize = resize(vit_data, size, anti_aliasing=True)
    H_resize = resize(H_data, size, anti_aliasing=True)
    L_resize = resize(L_data, size, anti_aliasing=True)

    vit_resize = vit_resize/vit_resize.max()
    H_resize = (H_resize - degfree)/(2*degfree)
    L_resize = (L_resize - degfree)/(2*degfree)

    vit_resize = vit_resize.reshape(vit_resize.shape[0],vit_resize.shape[1],1)
    H_resize = H_resize.reshape(1,H_resize.shape[0],H_resize.shape[1])
    L_resize = L_resize.reshape(1,L_resize.shape[0],L_resize.shape[1])
    
    spect_dat = np.vstack([H_resize, L_resize])

    model.eval()
    with torch.no_grad():

        prob = model.test(torch.Tensor([vit_resize,]), torch.Tensor([spect_dat,]), torch.Tensor([vit_stat,]) )

    return prob[0][0]


def main():
    #sftpath = "/home/pulsar/public_html/fscan/H1/weekly/H1Fscan_coherence/H1Fscan_coherence/"
    #channel = "H1_GDS-CALIB_STRAIN"
    #outpath = "/home/joseph.bayley/projects/soap_scripts/line_finding/imgs_test/"
    
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config-file', help='config file', type=str, required=False, default=None)

    parser.add_argument('-o', '--out-path', help='top level of output directories', type=str)
    parser.add_argument('-i', '--sft-path', help='path to sfts', type=str, required=False)
    parser.add_argument('-s', '--start-freq', help='start frequency for band', type=float, required=False)
    parser.add_argument('-e', '--end-freq', help='end frequency for band', type=float, required=False)
    parser.add_argument('-w', '--band-width', help='band_width', type=float, required=False, default = 0.1)
    parser.add_argument('--stride', help='stride, how many frequency bins to sum', type=int, required=False, default = 1)
    parser.add_argument('-r', '--obs-run', help='observing run', type=str, required=False)
    parser.add_argument('-l', '--lookup-dir', help='path to soap lookup tables', type=str, required=False)
    parser.add_argument('-sd', '--sub-dir', help='sub directory to save data', type=str, required=False)
    parser.add_argument('-cv', '--cnn-vitmap-model', help='filename of vitmap cnn model', type=str, required=False, default=None)
    parser.add_argument('-cs', '--cnn-spect-model', help='filename of spect cnn model', type=str, required=False, default=None)
    parser.add_argument('-cvs', '--cnn-vitmapstat-model', help='filename of vitmapstat cnn model', type=str, required=False, default=None)
    parser.add_argument('-ca', '--cnn-all-model', help='filename of vitmapstatspect cnn model', type=str, required=False, default=None)
    parser.add_argument('--force-overwrite', help='force overwrite tables and plots in output directory', action=argparse.BooleanOptionalAction)


    try:                                                     
        args = parser.parse_args()  
    except:  
        sys.exit(1)

    if args.cnn_vitmap_model is not None:

        #import tensorflow as tf
        #from keras import backend as K
        #from keras.models import load_model

        from skimage.transform import resize
        import torch

    if args.config_file is not None:
        cfg = SOAPConfig(args.config_file)
        #cfg = configparser.ConfigParser()
        #cfg.read(args.config_file)
    else:
        #outpath,minfreq,maxfreq,obs_run="O3",vitmapmodelfname=None, spectmodelfname = None, vitmapstatmodelfname = None, allmodelfname = None, sub_dir = "soap",
        cfg = {"output":{}, "data":{}, "input":{}, }

    if args.out_path:
        cfg["output"]["save_directory"] = args.out_path

    if args.sft_path:
        sftpath = args.sft_path.split(",")
        cfg["input"]["load_directory"] = sftpath

    run_soap_in_band(cfg, args.start_freq, args.end_freq, force_overwrite=args.force_overwrite)

if __name__ == "__main__":
    main()


import pkg_resources
import torch
import numpy as np
import configparser
import soapcw

def load_model(modelfile, device="cpu"):
    """
    Load in model weights from config and weights
    args
    ---------
    config: string
        path to the config ini file
    weights: string
        path to for the checkpointed weights of network

    returns
    ---------
    model_soap: torch.Module
        returns a torch model with trained weights
    config: SoapConfigParser
        returns the parsed config file for the model
    """

    pre_model = torch.load(modelfile, map_location=device).to(device)
    pre_model.device = device

    return model_soap

def load_model_from_config(config, weights, device="cpu"):
    """
    Load in model weights from config and weights
    args
    ---------
    config: string
        path to the config ini file
    weights: string
        path to for the checkpointed weights of network

    returns
    ---------
    model_soap: torch.Module
        returns a torch model with trained weights
    config: SoapConfigParser
        returns the parsed config file for the model
    """
    config = soapcw.soap_config_parser.SOAPConfig(config)

    pre_model_weights = torch.load(weights, map_location=device)

    model_soap = soapcw.neville.models.CVAE(
        input_dim=config["neville"]["input_dim"],
        par_dim=config["neville"]["num_predict_params"],
        latent_dim=config["neville"]["latent_dim"],
        fc_layers=config["neville"]["fc_layers"],
        conv_layers=config["neville"]["conv_layers"],
        inchannels=config["neville"]["n_channels"],
        fdim=config["neville"]["fdim"],
        dropout=config["neville"]["dropout"],
        dist_type=config["neville"]["dist_type"],
        device=device
        ).to(device)


    model_soap.load_state_dict(pre_model_weights["model_state_dict"])

    return model_soap, config



#!/usr/bin/env python

import sys
import os
import getopt
import time
import configparser 
import random
import numpy as np
import argparse
import logging
import subprocess


def create_dirs(dirs):
    for i in dirs:
        if not os.path.isdir(i):
            try: 
                os.makedirs(i)
            except:
                print("Could not create directory {}".format(i),file=sys.stderr)
                sys.exit(1)
    print("All directories exist")

def run_command_line(cl):
    """Run a string commandline as a subprocess, check for errors and return output."""

    logging.info('Executing: ' + cl)
    try:
        out = subprocess.check_output(cl,                       # what to run
                                      stderr=subprocess.STDOUT, # catch errors
                                      shell=True,               # proper environment etc
                                      universal_newlines=True   # properly display linebreaks in error/output printing
                                     )
    except subprocess.CalledProcessError as e:
        logging.error('Execution failed:')
        logging.error(e.output)
        print("-------failed {}".format(cl))
        out = None
        #raise
    os.system('\n')

    return(out)

def get_sft_filelist(file_list):
    """
    look through sftpaths to find all sft files
    """
    with open(file_list,"rb") as f:
        sftlist = f.readlines()

    return [fn.decode("utf-8").strip("\n") for fn in sftlist]

def get_filelist(out_path, sftdir, data_type):
    """ Creates a file with a list of all sfts in given directory"""
    filelist_fname = os.path.join(out_path, f"{data_type}_sft_list.txt")
    if os.path.isfile(filelist_fname):
        raise Exception("File already exists: Please delete this file if you want to rewrite")
    filelist_runstr = f"find {sftdir} -name '*.sft' -type f > {filelist_fname}"
    run_command_line(filelist_runstr)
    return filelist_fname

def split_sfts_list(config, detector, filelist_fname, output_dir):
    """Splits each of the sfts using lalapps split SFTs"""

    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    sft_list = get_sft_filelist(filelist_fname)

    #runstr = "find {} -name *.sft| sort -n -k4 -t '-'| xargs lalapps_splitSFTs -d {} -fs {} -fe {} -fb {} -fx {} -n {} -as 0 -- ".format(sft_path, detector, band_start, band_end, band_width, band_overlap, out_path)
    
    for fname in sft_list:
        try:
            runstr = "lalapps_splitSFTs -d {} -fs {} -fe {} -fb {} -fx {} -n {} -as 0 -- {}".format(detector, config["params"]["band_start"], config["params"]["band_end"], config["params"]["band_width"], config["params"]["band_overlap"], output_dir, fname)
            run_command_line(runstr)
        except:
            print(f"Split failed for SFT {fname}")


def write_subfile(config_file, params, output_dir, detector, filelist_fname):
    """Write a condor submit file to resubmit this file running the scripts to split the sfts"""
    dirs = [params["dirs"]["output_dir"]]
    for i in ["condor","condor/err","condor/log","condor/out"]:
        dirs.append(os.path.join(params["dirs"]["output_dir"],i))
        
    create_dirs(dirs)

    comment = "narrowband_{}_{}_{}_{}.sub".format(detector,params["params"]["obsrun"],params["params"]["band_start"],params["params"]["band_end"])
    sub_filename = os.path.join(*[params["dirs"]["output_dir"],"condor",comment])

    with open(sub_filename,'w') as f:
        f.write('# filename: {}\n'.format(sub_filename))
        f.write('universe = vanilla\n')
        f.write('executable = {}\n'.format(params["code"]["narrow_exec"]))
        #f.write('enviroment = ""\n')
        f.write('getenv  = True\n')
        f.write('log = {}/condor/log/{}_$(cluster).log\n'.format(params["dirs"]["output_dir"],comment))
        f.write('error = {}/condor/err/{}_$(cluster).err\n'.format(params["dirs"]["output_dir"],comment))
        f.write('output = {}/condor/out/{}_$(cluster).out\n'.format(params["dirs"]["output_dir"],comment))
        f.write('request_disk = 1000 \n')
        f.write(f'arguments = --config-file {config_file} --sft-filelist {filelist_fname} --output-dir {output_dir}\n')
        f.write(f'accounting_group = {params["params"]["accounting_group"]}\n')
        f.write('queue\n')
    
    print(time.time(), f"generated subfile - {sub_filename}")

    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-f', '--config-file', help='path to config file', type=str, required=True)
    parser.add_argument('-dag', '--make-dag', help='make the dag file for each detector', action="store_true")
    parser.add_argument('-d', '--detector', help='which detector to run on', type=str,  required=False)
    parser.add_argument('-s', '--sft-filelist', help='filelist of sfts', type=str,  required=False)
    parser.add_argument('-o', '--output-dir', help='output_directory for narrowbanded sfts', type=str,  required=False)


    args = parser.parse_args()  

    cp = configparser.ConfigParser()
    config_file = os.path.abspath(args.config_file)
    cp.read(config_file)

    if args.make_dag == True:
        for detector in cp["params"]["detectors"].split(","):
            detector = detector.strip()
            data_type = f'{detector}{cp["dirs"]["data_type"]}'
            sftpath = os.path.join(cp["dirs"]["sft_dir"], data_type)
            output_dir = os.path.join(cp["dirs"]["output_dir"], data_type)

            filelist_fname = get_filelist(cp["dirs"]["output_dir"], sftpath, data_type)

            write_subfile(config_file, cp, output_dir, detector, filelist_fname)

    else:
        split_sfts_list(cp, args.detector, args.sft_filelist, args.output_dir)

    



import lal
import numpy as np
import lalpulsar

class SFT:
    
    
    def __init__(self,nsft=None,tsft=None,delta_f=None,fmin=None,fmax=None,det_name=None,tstart=None):
        """
        Class to contain SFTs
        """
        self.sft = None
        self.norm_sft_power = None
        self.tstart = tstart
        self.nsft = nsft
        self.tsft = tsft
        self.epochs = None
        self.delta_f = delta_f
        self.det_name = det_name
        self.fmin = fmin
        self.fmax = fmax
        
        if self.delta_f is None and self.tsft is not None:
            self.delta_f = 1./self.tsft
            
            
    def sum_sft(self,sum_type="norm_sft_power",gap_val = 2,nsfts=48,remove_original=False):
        '''
        takes input sfts with tsft=1800s and returns sum of 48 sfts, i.e summing over a day
        and returns fraction of real data in each sum.
        args
        -------
        gap_val: double
            value used when there are gaps in the data (for calculating fraction)
        returns
        -----------
        data_av: array
            data summed over every day
        fraction: array
            fraction of each day which contained real data
        '''
        data = getattr(self,sum_type)
        data_av1 = []
        dutycycle = []
        sum_epochs = []
        sum_rng_med = []
        
        for i in np.arange(0,self.nsft,nsfts).astype(int):
            end = i+nsfts
            if end >= self.nsft:
                data = np.vstack([data,gap_val*np.ones((end - self.nsft, len(data[0])))])
            chunk = data[int(i):int(end)]
            av = np.nansum(chunk.T,axis=1)
            if np.isnan(gap_val):
                fr = 1. - len(chunk[:,0][np.isnan(chunk[:,0])])/float(nsfts)
            else:
                fr = 1. - len(chunk[:,0][chunk[:,0] == gap_val])/float(nsfts)
                        
            data_av1.append(av)
            dutycycle.append(fr)
            if hasattr(self, "rng_med"):
                sum_rng_med.append(np.nanmedian(self.rng_med[int(i):int(end)]))

            if self.epochs is not None:
                sum_epochs.append(self.epochs[int(i)] + 0.5*nsfts*self.tsft)

        setattr(self,"summed_{}".format(sum_type),np.array(data_av1))
        setattr(self,"summed_dutycycle",np.array(dutycycle))
        setattr(self,"summed_epochs",np.array(sum_epochs))
        if hasattr(self,"rng_med"):
            setattr(self, "summed_rng_med", np.array(sum_rng_med))

        
        if remove_original:
            delattr(self,"sft")

    def downsamp_frequency(self,data_type="norm_sft_power",stride=2,remove_original=False):
        '''
        downsample the frequency by taking the mean of n bins
        args
        -------
        stride: int
            number of bins to downsample by, default: 2, takes the mean of two bins
        '''
        data = getattr(self,data_type)

        if stride > self.nsft:
            raise Exception("Stride greater than length of SFT")
        if not hasattr(self,"nbins"):
            self.nbins = len(data[0])
            
        bins = np.arange(self.nbins)[::stride]
        if hasattr(self,"frequencies"):
            freqs = self.frequencies[bins] + self.delta_f
            setattr(self,"downsamp_frequencies",freqs)
            
        down_freq = np.array([np.sum(data[:,i:i+stride],axis=1) for i in bins]).T            
        setattr(self,"downsamp_{}".format(data_type),down_freq)
        
        if remove_original:
            delattr(self,data_type)

            
    def RngMed(self,med_width,sum_type="norm_sft_power"):
        """
        Calculate the running median for a set of data, data the same as the first point half of median window size is appended to the start and same to the end.
        
        Parameters
        -------------
        data: array
            1d array of data to find the running median for
        med_width: int
            the width of the running median window in sample points
        Returns
        ------------------
        med: array
           the running median of the data set with chosen window size
        """
        m = med_width // 2
        self.rng_med = np.zeros(np.shape(getattr(self,sum_type)))
        for ind,data in enumerate(getattr(self,sum_type)):
            data1 = np.insert(data,0,np.ones(m)*data[0])
            data1 = np.insert(data1,-1,np.ones(m)*data[-1])

            med = []
            for idx,pos in enumerate(data):
                med.append(np.nanmedian(data1[idx+m-m:idx+m+m+1]))
            self.rng_med[ind] = med
            del med


    def get_sh(self,pos='mean',inattr = None):
        """
        generate average psds as a function of time for data segment.
        args
        -----------
        data: array
            noise_data from function get_multi_likelihood (i.e data = [norm_data, epochs_data, tsft, noise_data]) 
            or similar so data[3] is noise data
        returns
        ----------
        psds: array
            array of average psd for band, where gaps are nans

        """
        psds = []
        if hasattr(self,"sft"):
            sh_dat = self.sft
        elif hasattr(self,"rng_med"):
            sh_dat = self.rng_med
        else:
            if inattr is None:
                print("No input attribute defined and not sft or running median")
            else:
                sh_dat = getattr(self,inattr)
        for i in range(len(sh_dat)):
            if pos == 'mean':
                shm = np.mean(sh_dat[i])
            elif pos == 'hmean':
                shm = st.hmean(sh_dat[i])
            elif pos == 'median':
                shm = np.median(sh_dat[i])
            else:
                j = pos
                shm = sh_dat[i][j]
            if shm == 2.0 or np.isnan(shm):
                psds.append(np.nan)
            else:
                psds.append(shm*2.0/self.tsft)

        return np.array(psds)
    
    def write_sft_files(self, output_path, narrowband = False, fmin = None, fmax = None):
        """
        Save and array of ffts to a set of LIGO sft files
        Parameters
        ---------------------
        output_path: string
            path where sfts will be saved
        narrowband: bool (optional: default - False)
            if True then will save one sft file containing all sfts, set between fmin and fmax, false saves a separate file for each sft
        fmin: float (optional)
            minimum frequency of sft
        fmax: float (optional)
            maximum frequency of sft
        

        """
        
        # if there are no epochs the define them here
        if self.epochs is None:
            self.epochs = np.linspace(self.start_time,self.start_time + self.nsft*self.tsft,self.nsft)
            
        if self.delta_f is None:
            self.delta_f = 1./self.tsft
        
        #set the length of the fft if fmin and max are set 
        #fft_length_index = int(self.tsft*1./(self.sample_frequency)/2.)
        if fmin is None:
            fmin = self.fmin
        if fmax is None:
            fmax = self.fmax
            
        if fmin < self.fmin:
            raise Exception("Please regenerate SFT down to this frequency")
        if fmax > self.fmax:
            raise Exception("Please regenerate SFT up to this frequency")
        
        fmin_index = int((fmin-self.fmin)*self.tsft)
        fmax_index = int((self.fmax-fmax)*self.tsft + 1)
        
        fft_length_index = fmax_index - fmin_index

        
        # if narrownaded sfts wanted then set up the sft vector for them to be appended to
        if narrowband:
            vct = lalpulsar.CreateSFTVector(int(self.nsft),int(fft_length_index ))
        
        
        #loop over the sfts and either save to files, or save to a narrowbanded sft
        sft_temp = None
        if self.nsft == 1:
            enm = enumerate([self.sft])
        else:
            enm = enumerate(self.sft)
        for t,fft in enm:
            if not narrowband:
                if sft_temp is None:
                    sft_temp = lal.CreateCOMPLEX8FrequencySeries(self.det_name, lal.LIGOTimeGPS(self.epochs[t]), fmin, 1./self.tsft, lal.SecondUnit, fft_length_index )
                sft_temp.epoch = self.epochs[t]
                sft_temp.data.data[:] = np.array(fft[fmin_index:fmax_index])
                sft_temp.deltaF = self.delta_f
                sft_temp.f0 = self.fmin
                
                filename_prefix = "{}/{}_{}SFT_T{}-{}_F{}_{}.sft".format(output_path, self.det_name,self.tsft, int(self.epochs[t]) ,int(self.epochs[-1]-self.epochs[0]), fmin, fmax)
                lalpulsar.WriteSFT2file(sft_temp, filename_prefix, "")
            elif narrowband:
                vct.data[t].data.data[:] = np.array(fft[fmin_index:fmax_index]).astype("complex64")
                vct.data[t].deltaF = self.delta_f
                vct.data[t].f0 = self.fmin
                vct.data[t].epoch = self.epochs[t]
                if self.det_name is not None:
                    vct.data[t].name = self.det_name
                else:
                    vct.data[t].name = "NoDet"
            
        if narrowband: 
            filename_prefix = "{}/{}-{}_{}SFT_T{}-{}_F{}_{}.sft".format(output_path,self.det_name,self.nsft,self.tsft, int(self.epochs[0]) ,int(self.epochs[-1]-self.epochs[0]), fmin, fmax)
            lalpulsar.WriteSFTVector2NamedFile(vct, filename_prefix, "")
            
